import os
import struct
import numpy as np
from NEDAS.utils.conversion import type_dic, type_size, t2h, h2t, dt1h, ensure_list
from NEDAS.utils.parallel import distribute_tasks, bcast_by_root
from NEDAS.utils.progress import progress_bar

class State:
    """
    The State class manages the state variables for the assimilation system.
    
    The analysis is performed on a regular grid.
    
    The entire state has dimensions: member, variable, time,  z,  y,  x
    indexed by: mem_id,        v,    t,  k,  j,  i
    with size:   nens,       nv,   nt, nz, ny, nx

    To parallelize workload, we group the dimensions into 3 indices:
    
    mem_id indexes the ensemble members
    
    rec_id indexes the uniq 2D fields with (v, t, k), since nz and nt may vary
    for different variables, we stack these dimensions in the 'record'
    dimension with size nrec
    
    par_id indexes the spatial partitions, which are subset of the 2D grid
    given by (ist, ied, di, jst, jed, dj), for a complete field fld[j,i]
    the processor with par_id stores fld[ist:ied:di, jst:jed:dj] locally.

    The entire state is distributed across the memory of many processors,
    at any moment, a processor only stores a subset of state in its memory:
    either having all the mem_id,rec_id but only a subset of par_id (we call this
    ensemble-complete), or having all the par_id but a subset of mem_id,rec_id
    (we call this field-complete).
    It is easier to perform i/o and pre/post processing on field-complete state,
    while easier to run assimilation algorithms with ensemble-complete state.
    """
    def __init__(self, c):
        self.analysis_dir = c.analysis_dir(c.time, c.iter)
        self.prior_file = os.path.join(self.analysis_dir, 'prior_state.bin')
        self.prior_mean_file = os.path.join(self.analysis_dir, 'prior_mean_state.bin')
        self.post_file = os.path.join(self.analysis_dir, 'post_state.bin')
        self.post_mean_file = os.path.join(self.analysis_dir, 'post_mean_state.bin')
        self.z_coords_file = os.path.join(self.analysis_dir, 'z_coords.bin')

        self.info = bcast_by_root(c.comm)(self.parse_state_info)(c)

        self.mem_list, self.rec_list = bcast_by_root(c.comm)(self.distribute_state_tasks)(c)

        self.partitions = {}  ##will be setup by assimilator.partition_grid()
        self.par_list = {}

        self.fields_prior = {}  ##will be created by prepare_obs()
        self.z_fields = {}
        self.state_prior = {}   ##will be created by transpose_to_ensemble_complete()
        self.z_state = {}
        self.state_post = {}    ##will be created by assimilator.assimilate()
        self.fields_post = {}   ##will be created by transpose_to_field_complete()
        self.data = {}  ##will be created by pack_state_data, for use in assimilate()

    def prepare_state(self, c):
        """
        Main method from state object to be called in the analysis script
        """
        self.fields_prior, self.z_fields = self.collect_fields_from_restartfiles(c)
        self.output_z_coords(c)

    def parse_state_info(self, c) -> dict:
        """
        Parses info for the nrec fields in the state.

        Args:
            c (Config): config obj

        Returns:
            dict: A dictionary with some dimensions and list of unique field records
        """
        info = {}
        info['size'] = 0
        info['shape'] = c.grid.x.shape
        info['fields'] = {}
        info['scalars'] = {}
        rec_id = 0   ##record id for a 2D field
        pos = 0      ##seek position for rec
        variables = set()
        err_types = set()

        ##loop through variables in state_def
        for vrec in ensure_list(c.state_def):
            vname = vrec['name']
            variables.add(vname)
            err_types.add(vrec['err_type'])

            if vrec['var_type'] == 'field':
                ##this is a state variable 'field' with dimensions t, z, y, x
                ##some properties of the variable is defined in its source module
                src = c.models[vrec['model_src']]
                assert vname in src.variables, 'variable '+vname+' not defined in '+vrec['model_src']+' Model.variables'

                #now go through time and zlevels to form a uniq field record
                for time in c.time + np.array(c.state_time_steps)*dt1h:
                    for k in src.variables[vname]['levels']:
                        rec = { 'name': vname,
                                'model_src': vrec['model_src'],
                                'dtype': src.variables[vname]['dtype'],
                                'is_vector': src.variables[vname]['is_vector'],
                                'units': src.variables[vname]['units'],
                                'err_type': vrec['err_type'],
                                'time': time,
                                'dt': c.state_time_scale,
                                'k': k,
                                'pos': pos, }
                        info['fields'][rec_id] = rec

                        ##update seek position
                        nv = 2 if rec['is_vector'] else 1
                        fld_size = np.sum((~c.grid.mask).astype(int))
                        pos += nv * fld_size * type_size[rec['dtype']]
                        rec_id += 1

            elif vrec['var_type'] == 'scalar':
                pass

            else:
                raise NotImplementedError(f"{vrec['var_type']} is not supported in the state vector.")

        if c.debug:
            print(f"number of ensemble members, nens={c.nens}", flush=True)
            print(f"number of unique field records, nrec={len(info['fields'])}", flush=True)
            print(f"variables: {variables}", flush=True)

        info['size'] = pos ##size of a complete state (fields) for 1 memeber
        info['variables'] = list(variables)
        info['err_types'] = list(err_types)

        return info

    def write_state_info(self, binfile: str):
        """
        Write state_info to a .dat file accompanying the .bin file

        Args:
            binfile (str): File path for the .bin file
        """
        with open(binfile.replace('.bin','.dat'), 'wt') as f:
            ##first line: grid dimension
            if len(self.info['shape']) == 1:
                f.write('{}\n'.format(self.info['shape'][0]))
            else:
                f.write('{} {}\n'.format(self.info['shape'][0], self.info['shape'][1]))

            ##second line: total size of the state
            f.write('{}\n'.format(self.info['size']))

            ##followed by nfield lines: each for a field record
            for i, rec in self.info['fields'].items():
                name = rec['name']
                model_src = rec['model_src']
                dtype = rec['dtype']
                is_vector = int(rec['is_vector'])
                units = rec['units']
                err_type = rec['err_type']
                time = t2h(rec['time'])
                dt = rec['dt']
                k = rec['k']
                pos = rec['pos']
                f.write('{} {} {} {} {} {} {} {} {} {}\n'.format(name, model_src, dtype, is_vector, units, err_type, time, dt, k, pos))

    def read_state_info(self, binfile: str) -> dict:
        """
        Read .dat file accompanying the .bin file and obtain state_info

        Args:
            binfile (str): File path for the .bin file

        Returns:
            dict: state info dictionary
        """
        with open(binfile.replace('.bin','.dat'), 'r') as f:
            lines = f.readlines()
            info = {}

            ss = lines[0].split()
            if len(ss)==1:
                info['shape'] = (int(ss),)
            else:
                info['shape'] = (int(ss[0]), int(ss[1]))

            info['size'] = int(lines[1])

            ##records for uniq fields
            info['fields'] = {}
            rec_id = 0
            for lin in lines[2:]:
                ss = lin.split()
                rec = {'name': ss[0],
                    'model_src': ss[1],
                    'dtype': ss[2],
                    'is_vector': bool(int(ss[3])),
                    'units': ss[4],
                    'err_type': ss[5],
                    'time': h2t(np.float32(ss[6])),
                    'dt': np.float32(ss[7]),
                    'k': np.float32(ss[8]),
                    'pos': int(ss[9]), }
                info['fields'][rec_id] = rec
                rec_id += 1
            return info

    def write_field(self, binfile: str, mask: np.ndarray, mem_id: int, rec_id: int, fld: np.ndarray) -> None:
        """
        Write a field to a binary file

        Args:
            binfile (str): File path for the .bin file
            mask (np.ndarray):  Array of bool with same shape as grid.x.
                True if the grid point is masked (for example land grid point in ocean models).
                The masked points will not be stored in the binfile to reduce disk usage.
            mem_id (int): Member index, from 0 to nens-1
            rec_id (int): Field record index, info['fields'][rec_id] gives the record information
            fld (np.ndarray):  The field to be written to the file
        """
        rec = self.info['fields'][rec_id]

        fld_shape = (2,)+self.info['shape'] if rec['is_vector'] else self.info['shape']
        assert fld.shape == fld_shape, f'fld shape incorrect: expected {fld_shape}, got {fld.shape}'

        if rec['is_vector']:
            fld_ = fld[:, ~mask].flatten()
        else:
            fld_ = fld[~mask]

        with open(binfile, 'r+b') as f:
            f.seek(mem_id*self.info['size'] + rec['pos'])
            f.write(struct.pack(fld_.size*type_dic[rec['dtype']], *fld_))

    def read_field(self, binfile: str, mask: np.ndarray, mem_id: int, rec_id: int) -> np.ndarray:
        """
        Read a field from a binary file

        Args:
        binfile (str): File path for the .bin file
        mask (np.ndarray): Mask of grid points.
        mem_id (int): Member index from 0 to nens-1
        rec_id (int): Field record index, info['fields'][rec_id] gives the record information

        Returns:
            np.ndarray: The field read from the file
        """
        rec = self.info['fields'][rec_id]

        nv = 2 if rec['is_vector'] else 1
        fld_shape = (2,)+self.info['shape'] if rec['is_vector'] else self.info['shape']
        fld_size = np.sum((~mask).astype(int))

        with open(binfile, 'rb') as f:
            f.seek(mem_id*self.info['size'] + rec['pos'])
            fld_ = np.array(struct.unpack((nv*fld_size*type_dic[rec['dtype']]),
                            f.read(nv*fld_size*type_size[rec['dtype']])))
            fld = np.full(fld_shape, np.nan)
            if rec['is_vector']:
                fld[:, ~mask] = fld_.reshape((2, -1))
            else:
                fld[~mask] = fld_
            return fld

    def distribute_state_tasks(self, c):
        """
        Distribute mem_id and rec_id across processors

        Args:
        c (Config): config obj

        Returns:
            dict: mem_list {pid_mem: list[mem_id]}
            dict: rec_list {pid_rec: list[rec_id]}
        """
        ##list of mem_id as tasks
        mem_list = distribute_tasks(c.comm_mem, [m for m in range(c.nens)])

        ##list rec_id as tasks
        rec_list_full = [i for i in self.info['fields'].keys()]
        rec_size = np.array([2 if r['is_vector'] else 1 for i,r in self.info['fields'].items()])
        rec_list = distribute_tasks(c.comm_rec, rec_list_full, rec_size)

        return mem_list, rec_list

    def output_state(self, c, fields, state_file, mem_id_out=None, rec_id_out=None):
        """
        Parallel output the fields to the binary state_file

        Args:
            c (Config): config obj
            fields (dict): the locally-stored field-complete fields for output, [(mem_id, rec_id), fld]
            state_file (str): path to the output binary file.
            mem_id_out (int, optional): member id to be output, if None all available ids will output.
            rec_id_out (int, optional): record id to be output, if None all available ids will output.
        """
        c.print_1p('>>> save state to '+state_file+'\n')

        if c.pid == 0:
            ##if file doesn't exist, create the file
            open(state_file, 'wb')
            ##write state_info to the accompanying .dat file
            self.write_state_info(state_file)
        c.comm.Barrier()

        nm = len(self.mem_list[c.pid_mem])
        nr = len(self.rec_list[c.pid_rec])
        for m, mem_id in enumerate(self.mem_list[c.pid_mem]):
            if mem_id_out is not None and mem_id != mem_id_out:
                continue
            for r, rec_id in enumerate(self.rec_list[c.pid_rec]):
                if rec_id_out is not None and rec_id != rec_id_out:
                    continue

                if c.debug:
                    rec = self.info['fields'][rec_id]
                    print(f"PID {c.pid:4}: saving field: mem{mem_id+1:03} '{rec['name']:20}' {rec['time']} k={rec['k']}", flush=True)
                else:
                    c.print_1p(progress_bar(m*nr+r, nm*nr))

                ##get the field record for output
                fld = fields[mem_id, rec_id]

                ##write the data to binary file
                self.write_field(state_file, c.grid.mask, mem_id, rec_id, fld)

        c.comm.Barrier()
        c.print_1p(' done.\n')

    def output_ens_mean(self, c, fields, mean_file):
        """
        Compute ensemble mean of a field stored distributively on all pid_mem
        collect means on pid_mem=0, and output to mean_file

        Args:
            c (Config): config obj
            fields (dict): the locally stored field-complete fields for output
            mean_file (str): path to the output binary file for the ensemble mean
        """
        c.print_1p('>>> compute ensemble mean, save to '+mean_file+'\n')

        if c.pid == 0:
            ##if file doesn't exist, create the file, write state_info
            open(mean_file, 'wb')
            self.write_state_info(mean_file)
        c.comm.Barrier()

        for r, rec_id in enumerate(self.rec_list[c.pid_rec]):
            rec = self.info['fields'][rec_id]
            if c.debug:
                print(f"PID {c.pid:4}: saving mean field '{rec['name']:20}' {rec['time']} k={rec['k']}", flush=True)
            else:
                c.print_1p(progress_bar(r, len(self.rec_list[c.pid_rec])))

            ##initialize a zero field with right dimensions for rec_id
            fld_shape = (2,)+self.info['shape'] if rec['is_vector'] else self.info['shape']
            sum_fld_pid = np.zeros(fld_shape)

            ##sum over all fields locally stored on pid
            for mem_id in self.mem_list[c.pid_mem]:
                sum_fld_pid += fields[mem_id, rec_id]

            ##sum over all field sums on different pids together to get the total sum
            ##TODO:reduce is expensive if only part of pid holds state in memory
            sum_fld = c.comm_mem.reduce(sum_fld_pid, root=0)

            if c.pid_mem == 0:
                mean_fld = sum_fld / c.nens
                self.write_field(mean_file, c.grid.mask, 0, rec_id, mean_fld)

        c.comm.Barrier()
        c.print_1p(' done.\n')

    def output_z_coords(self, c):
        ##topaz uses the first ensemble member z coords as the reference z for obs
        ##include this here for backward compatibility
        ##there is no need for choosing which member also, just use the first one
        if c.z_coords_from == 'member':
            self.output_state(c, self.z_fields, self.z_coords_file, mem_id_out=0)

        ##we use by default the ensemble mean z coords as the reference z for obs
        if c.z_coords_from == 'mean':
            self.output_ens_mean(c, self.z_fields, self.z_coords_file)

    def collect_fields_from_restartfiles(self, c):
        """
        Collects fields from model restart files, convert them to the analysis grid,
        preprocess (coarse-graining etc), save to fields[mem_id, rec_id] pointing to the uniq fields

        Args:
            c (Config): config object

        Returns:
            dict: fields dictionary [(mem_id, rec_id), fld]
                where fld is np.array defined on c.grid, it's one of the state variable field
            dict: z_fields dictionary [(mem_id, rec_id), zfld]
                where zfld is same shape as fld, it's he z coordinates corresponding to each field
        """
        pid_mem_show = [p for p,lst in self.mem_list.items() if len(lst)>0][0]
        pid_rec_show = [p for p,lst in self.rec_list.items() if len(lst)>0][0]
        c.pid_show =  pid_rec_show * c.nproc_mem + pid_mem_show

        ##pid_show has some workload, it will print progress message
        c.print_1p('>>> prepare state by reading fields from model restart\n')
        fields = {}
        z_fields = {}

        ##process the fields, each proc gets its own workload as a subset of
        ##mem_id,rec_id; all pid goes through their own task list simultaneously
        nm = len(self.mem_list[c.pid_mem])
        nr = len(self.rec_list[c.pid_rec])

        for m, mem_id in enumerate(self.mem_list[c.pid_mem]):
            for r, rec_id in enumerate(self.rec_list[c.pid_rec]):
                rec = self.info['fields'][rec_id]

                if c.debug:
                    print(f"PID {c.pid:4}: prepare_state mem{mem_id+1:03} '{rec['name']:20}' {rec['time']} k={rec['k']}", flush=True)
                else:
                    c.print_1p(progress_bar(m*nr+r, nm*nr))

                ##directory storing model output
                path = c.forecast_dir(rec['time'], rec['model_src'])

                ##the model object for handling this variable
                model = c.models[rec['model_src']]

                ##read field from restart file
                model.read_grid(path=path, member=mem_id, **rec)
                var = model.read_var(path=path, member=mem_id, **rec)

                model.grid.set_destination_grid(c.grid)
                fld = model.grid.convert(var, is_vector=rec['is_vector'], method='linear', coarse_grain=True)

                if rec['is_vector']:
                    fld[:, c.grid.mask] = np.nan
                else:
                    fld[c.grid.mask] = np.nan

                ##misc. transform can be added here
                for transform_func in c.transform_funcs:
                    fld = transform_func.forward_state(c, rec, fld)

                ##save field to dict
                fields[mem_id, rec_id] = fld

                ##read z_coords for the field
                ##only need to generate the uniq z coords, store in bank
                zvar = model.z_coords(path=path, member=mem_id, **rec)
                z = model.grid.convert(zvar, is_vector=False, method='linear', coarse_grain=True)
                if rec['is_vector']:
                    z_fields[mem_id, rec_id] = np.array([z, z])
                else:
                    z_fields[mem_id, rec_id] = z

        c.comm.Barrier()
        c.print_1p(' done.\n')

        ##additonal output of debugging
        if c.debug:
            np.save(os.path.join(self.analysis_dir, f'fields_prior.{c.pid_mem}.{c.pid_rec}.npy'), fields)

        return fields, z_fields

    def collect_scalar_variables(self, c):
        pass
        # TODO: implement scalars here for simultaneous state parameter estimation (SSPE)

    def pack_field_chunk(self, c, fld, is_vector, dst_pid):
        fld_chk = {}
        for par_id in self.par_list[dst_pid]:
            if len(c.grid.x.shape) == 2:
                ##slice for this par_id
                istart,iend,di,jstart,jend,dj = self.partitions[par_id]
                ##save the unmasked points in slice to fld_chk for this par_id
                mask_chk = c.grid.mask[jstart:jend:dj, istart:iend:di]
                if is_vector:
                    fld_chk[par_id] = fld[:, jstart:jend:dj, istart:iend:di][:, ~mask_chk]
                else:
                    fld_chk[par_id] = fld[jstart:jend:dj, istart:iend:di][~mask_chk]
            else:
                inds = self.partitions[par_id]
                mask_chk = c.grid.mask[inds]
                if is_vector:
                    fld_chk[par_id] = fld[:, inds][:, ~mask_chk]
                else:
                    fld_chk[par_id] = fld[inds][~mask_chk]
        return fld_chk

    def unpack_field_chunk(self, c, fld, fld_chk, src_pid):
        for par_id in self.par_list[src_pid]:
            if len(c.grid.x.shape) == 2:
                istart,iend,di,jstart,jend,dj = self.partitions[par_id]
                mask_chk = c.grid.mask[jstart:jend:dj, istart:iend:di]
                fld[..., jstart:jend:dj, istart:iend:di][..., ~mask_chk] = fld_chk[par_id]
            else:
                inds = self.partitions[par_id]
                mask_chk = c.grid.mask[inds]
                fld[..., inds][..., ~mask_chk] = fld_chk[par_id]

    def transpose_to_ensemble_complete(self, c, fields):
        """
        Send chunks of field owned by a pid to other pid
        so that the field-complete fields get transposed into ensemble-complete state
        with keys (mem_id, rec_id) pointing to the partition in par_list

        Args:
            c (Config): config obj
            fields (dict): The locally stored field-complete fields with subset of mem_id,rec_id

        Returns:
            dict: The locally stored ensemble-complete field chunks on partitions, dict[(mem_id, rec_id), dict[par_id, fld_chk]]
        """
        c.print_1p('transpose field-complete to ensemble-complete\n')
        state = {}

        nr = len(self.rec_list[c.pid_rec])
        for r, rec_id in enumerate(self.rec_list[c.pid_rec]):

            ##all pid goes through their own mem_list simultaneously
            nm_max = np.max([len(lst) for p,lst in self.mem_list.items()])
            for m in range(nm_max):
                if c.debug:
                    if m < len(self.mem_list[c.pid_mem]):
                        mem_id = self.mem_list[c.pid_mem][m]
                        print(f"PID {c.pid:4}: transposing field: mem{mem_id+1:03} rec{rec_id}")
                    else:
                        print(f"PID {c.pid:4}: transposing field: waiting")
                else:
                    c.print_1p(progress_bar(r*nm_max+m, nr*nm_max))

                ##prepare the fld for sending if not at the end of mem_list
                if m < len(self.mem_list[c.pid_mem]):
                    mem_id = self.mem_list[c.pid_mem][m]
                    rec = self.info['fields'][rec_id]
                    fld = fields[mem_id, rec_id].copy()

                ## - for each source pid_mem (src_pid) with fields[mem_id, rec_id],
                ##   send chunk of fld to destination pid_mem (dst_pid) with its partition in par_list
                ## - every pid needs to send/recv to/from every pid, so we use cyclic
                ##   coreography here to prevent deadlock

                ## 1) receive fld_chk from src_pid, for src_pid<pid first
                for src_pid in np.arange(0, c.pid_mem):
                    if m < len(self.mem_list[src_pid]):
                        src_mem_id = self.mem_list[src_pid][m]
                        state[src_mem_id, rec_id] = c.comm_mem.recv(source=src_pid, tag=m)

                ## 2) send my fld chunk to a list of dst_pid, send to dst_pid>=pid first
                ##    because they wait to receive before able to send their own stuff;
                ##    when finished with dst_pid>=pid, cycle back to send to dst_pid<pid,
                ##    i.e., dst_pid list = [pid, pid+1, ..., nproc-1, 0, 1, ..., pid-1]
                if m < len(self.mem_list[c.pid_mem]):
                    for dst_pid in np.mod(np.arange(c.nproc_mem)+c.pid_mem, c.nproc_mem):
                        fld_chk = self.pack_field_chunk(c, fld, rec['is_vector'], dst_pid)
                        if dst_pid == c.pid_mem:
                            ##same pid, so just write to state
                            state[mem_id, rec_id] = fld_chk
                        else:
                            ##send fld_chk to dst_pid's state
                            c.comm_mem.send(fld_chk, dest=dst_pid, tag=m)

                ## 3) finish receiving fld_chk from src_pid, for src_pid>pid now
                for src_pid in np.arange(c.pid_mem+1, c.nproc_mem):
                    if m < len(self.mem_list[src_pid]):
                        src_mem_id = self.mem_list[src_pid][m]
                        state[src_mem_id, rec_id] = c.comm_mem.recv(source=src_pid, tag=m)
        c.comm.Barrier()
        c.print_1p(' done.\n')
        return state

    def transpose_to_field_complete(self, c, state):
        """
        Transposes back the state to field-complete fields

        Args:
            c (Config): config obj
            state (dict): the locally stored ensemble-complete field chunks for subset of par_id

        Returns:
            dict: the locally stored field-complete fields for subset of mem_id,rec_id.
        """
        c.print_1p('transpose ensemble-complete to field-complete\n')
        fields = {}

        ##all pid goes through their own task list simultaneously
        nr = len(self.rec_list[c.pid_rec])
        for r, rec_id in enumerate(self.rec_list[c.pid_rec]):

            ##all pid goes through their own mem_list simultaneously
            nm_max = np.max([len(lst) for p,lst in self.mem_list.items()])

            for m in range(nm_max):
                if c.debug:
                    if m < len(self.mem_list[c.pid_mem]):
                        mem_id = self.mem_list[c.pid_mem][m]
                        print(f"PID {c.pid:4}: transposing field: mem{mem_id+1:03} rec{rec_id}")
                    else:
                        print(f"PID {c.pid:4}: transposing field: waiting")
                else:
                    c.print_1p(progress_bar(r*nm_max+m, nr*nm_max))

                ##prepare an empty fld for receiving if not at the end of mem_list
                if m < len(self.mem_list[c.pid_mem]):
                    mem_id = self.mem_list[c.pid_mem][m]
                    rec = self.info['fields'][rec_id]
                    if rec['is_vector']:
                        fld = np.full((2,)+c.grid.x.shape, np.nan)
                    else:
                        fld = np.full(c.grid.x.shape, np.nan)
                    fields[mem_id, rec_id] = fld

                ##this is just the reverse of transpose_field_to_state
                ## we take the exact steps, but swap send and recv operations here
                ##
                ## 1) send my fld_chk to dst_pid, for dst_pid<pid first
                for dst_pid in np.arange(0, c.pid_mem):
                    if m < len(self.mem_list[dst_pid]):
                        dst_mem_id = self.mem_list[dst_pid][m]
                        c.comm_mem.send(state[dst_mem_id, rec_id], dest=dst_pid, tag=m)
                        del state[dst_mem_id, rec_id]   ##free up memory

                ## 2) receive fld_chk from a list of src_pid, from src_pid>=pid first
                ##    because they wait to send stuff before able to receive themselves,
                ##    cycle back to receive from src_pid<pid then.
                if m < len(self.mem_list[c.pid_mem]):
                    for src_pid in np.mod(np.arange(c.nproc_mem)+c.pid_mem, c.nproc_mem):
                        if src_pid == c.pid_mem:
                            ##same pid, so just copy fld_chk from state
                            fld_chk = state[mem_id, rec_id].copy()
                        else:
                            ##receive fld_chk from src_pid's state
                            fld_chk = c.comm_mem.recv(source=src_pid, tag=m)

                        ##unpack the fld_chk to form a complete field
                        self.unpack_field_chunk(c, fld, fld_chk, src_pid)

                ## 3) finish sending fld_chk to dst_pid, for dst_pid>pid now
                for dst_pid in np.arange(c.pid_mem+1, c.nproc_mem):
                    if m < len(self.mem_list[dst_pid]):
                        dst_mem_id = self.mem_list[dst_pid][m]
                        c.comm_mem.send(state[dst_mem_id, rec_id], dest=dst_pid, tag=m)
                        del state[dst_mem_id, rec_id]   ##free up memory
        c.comm.Barrier()
        c.print_1p(' done.\n')
        return fields

    def pack_local_state_data(self, c, par_id, state_prior, z_state):
        """pack state dict into arrays to be more easily handled by jitted funcs"""
        data = {}

        ##x,y coordinates for local state variables on pid
        if len(c.grid.x.shape) == 2:  ##regular grid
            ist,ied,di,jst,jed,dj = self.partitions[par_id]
            msk = c.grid.mask[jst:jed:dj, ist:ied:di]
            data['x'] = c.grid.x[jst:jed:dj, ist:ied:di][~msk]
            data['y'] = c.grid.y[jst:jed:dj, ist:ied:di][~msk]

        else:
            inds = self.partitions[par_id]
            msk = c.grid.mask[inds]
            data['x'] = c.grid.x[inds][~msk]
            data['y'] = c.grid.y[inds][~msk]

        data['field_ids'] = []
        for rec_id in self.rec_list[c.pid_rec]:
            rec = self.info['fields'][rec_id]
            v_list = [0, 1] if rec['is_vector'] else [None]
            for v in v_list:
                data['field_ids'].append((rec_id, v))

        nfld = len(data['field_ids'])
        nloc = len(data['x'])
        data['t'] = np.full(nfld, np.nan)
        data['z'] = np.zeros((nfld, nloc))
        data['var_id'] = np.full(nfld, 0)
        data['err_type'] = np.full(nfld, 0)
        data['state_prior'] = np.full((c.nens, nfld, nloc), np.nan)
        for n in range(nfld):
            rec_id, v = data['field_ids'][n]
            rec = self.info['fields'][rec_id]
            data['t'][n] = t2h(rec['time'])
            data['err_type'][n] = self.info['err_types'].index(rec['err_type'])
            data['var_id'][n] = self.info['variables'].index(rec['name'])
            for m in range(c.nens):
                data['z'][n, :] += np.squeeze(z_state[m, rec_id][par_id][v, :]).astype(np.float32) / c.nens  ##ens mean z
                data['state_prior'][m, n, :] = np.squeeze(state_prior[m, rec_id][par_id][v, :].copy())

        return data

    def unpack_local_state_data(self, c, par_id, state_prior, data):
        """unpack data and write back to the state dict"""
        nfld = len(data['field_ids'])
        nloc = len(data['x'])

        for m in range(c.nens):
            for n in range(nfld):
                rec_id, v = data['field_ids'][n]
                state_prior[m, rec_id][par_id][v, :] = data['state_prior'][m, n, :]



import importlib
from NEDAS.config import Config
from .base import Updator

registry = {
    'additive': 'AdditiveUpdator',
    'alignment': 'AlignmentUpdator',
    'alignment_interp': 'AdditiveUpdator',
}

def get_updator(c: Config) -> Updator:
    """
    Get the correct Updator subclass instance based on the configuration.

    Args:
        c (Config): Configuration object.

    Returns:
        Updator: Corresponding Updator subclass instance.
    """
    if not hasattr(c, 'updator_def'):
        raise AttributeError("'updator_def' missing in configuration")
    if not isinstance(c.updator_def, dict):
        c.updator_def = {}
    if 'type' not in c.updator_def.keys():
        c.updator_def['type'] = 'additive'
    updator_type = c.updator_def['type'].lower()

    if updator_type not in registry:
        raise NotImplementedError(f"updator type '{updator_type}' is not implemented")

    ##TODO: last scale component doesn't need alignment, find a better general logic
    if c.iter == c.niter-1:
        updator_type = 'additive' 
    module = importlib.import_module('NEDAS.assim_tools.updators.'+updator_type)
    UpdatorClass = getattr(module, registry[updator_type])

    return UpdatorClass(c)



import os
import configparser
from NEDAS.utils.conversion import dt1h
from NEDAS.models.nextsim.dg import restart, forcing

def make_namelist(file_options:dict, model_config_file:str, ens_dir='.', **kwargs):
    ens_mem_id:int = kwargs['member'] + 1  ##TODO: member could be None for deterministic runs
    time = kwargs['time']
    forecast_period = kwargs['forecast_period']
    next_time = time + forecast_period * dt1h
    time_start = kwargs['time_start']

    # read the config file
    model_config = configparser.ConfigParser()
    model_config.optionxform = str
    model_config.read(model_config_file)

    ##change the restart file name
    file_options_restart = file_options['restart']
    fname_restart:str = restart.get_restart_filename(file_options_restart, ens_mem_id, time)
    model_config['model']['init_file'] = os.path.basename(fname_restart)
    model_config['model']['start'] = time.strftime("%Y-%m-%dT%H:%M:%SZ")
    model_config['model']['stop'] = next_time.strftime("%Y-%m-%dT%H:%M:%SZ")
    model_config['ConfigOutput']['start'] = time.strftime("%Y-%m-%dT%H:%M:%SZ")
    # changing the forcing file in ERA5Atmosphere
    file_options_forcing:dict[str, str] = file_options['forcing']
    fname_atmos_forcing = forcing.get_forcing_filename(file_options_forcing['atmosphere'],
                                                        1, time)
    fname_atmos_forcing = os.path.basename(fname_atmos_forcing)
    model_config['ERA5Atmosphere']['file'] = fname_atmos_forcing
    # changing the forcing file in ERA5Atmosphere
    fname_ocn_forcing = forcing.get_forcing_filename(file_options_forcing['ocean'],
                                                        1, time)
    fname_ocn_forcing = os.path.basename(fname_ocn_forcing)
    model_config['TOPAZOcean']['file'] = fname_ocn_forcing

    # dump the config to new file
    config_file = os.path.join(ens_dir, 'nextsim.cfg')
    with open(config_file, 'w') as configfile:
        model_config.write(configfile)




"""
"""
from datetime import datetime
import cftime
import netCDF4
import numpy as np

def copy_group(src_group, dst_group, time_index, time_dimension_name, time) -> None:
    """
    Recursively copy a netCDF group, including dimensions, variables, and attributes.
    """
    # Copy global (or group) attributes
    dst_group.setncatts(src_group.__dict__)

    # Copy dimensions
    for name, dimension in src_group.dimensions.items():
        dst_group.createDimension(name, len(dimension) if not dimension.isunlimited() else None)

    # Copy variables
    for name, variable in src_group.variables.items():
        # Create the variable in the destination group
        dst_var = dst_group.createVariable(name, variable.datatype, variable.dimensions)
        dst_var.setncatts({k: variable.getncattr(k) for k in variable.ncattrs()})

        if len(variable.dimensions) == 0:
            if name == 'formatted':
                dst_var[0] = time.strftime(dst_var.format)
            if name == 'time':
                dst_var[0] = cftime.date2num(time, dst_var.units)
            continue
        # Copy the data, considering the time slice if relevant
        if time_dimension_name in variable.dimensions:
            # For other time-dependent variables, copy the time slice
            slice_spec = [time_index if dim == time_dimension_name else slice(None) for dim in variable.dimensions]
            dst_var[:] = variable[tuple(slice_spec)]
        else:
            # For time-independent variables, copy all data
            dst_var[:] = variable[:]

    # Recursively copy groups
    for group_name in src_group.groups:
        src_subgroup = src_group.groups[group_name]
        dst_subgroup = dst_group.createGroup(group_name)
        copy_group(src_subgroup, dst_subgroup, time_index, time_dimension_name, time)

def copy_time_sliced_nc_file(source_file:str, target_file,
                             time_index:np.ndarray,
                             time_varname:str,
                             time:datetime) -> None:
    # Open the source netCDF file in read mode
    with netCDF4.Dataset(source_file, 'r') as src:
        # Open the target netCDF file in write mode
        with netCDF4.Dataset(target_file, 'w') as dst:
            # Copy the root group (dimensions, variables, and groups)
            time_dimension_name = src[time_varname].dimensions[0]
            copy_group(src, dst, time_index, time_dimension_name, time)
            dst.sync()

if __name__ == '__main__':
    copy_time_sliced_nc_file('/bettik/yumengch-ext/DATA/25km_NH.TOPAZ4_2010-01-01_2011-01-01.nc',
                             'test.nc', np.arange(2), time_dimension_name='time', )


import importlib
from typing import Any

registry = {
    'offline_filter': 'OfflineFilterAnalysisScheme',
}

def get_analysis_scheme(c) -> Any:
    """
    Factory function to get the correct analysis scheme instance.

    Args:
        c (Config): Configuration.

    Returns:
        The analysis scheme class instance.
    """
    if not hasattr(c, 'analysis_scheme'):
        raise KeyError("Configuration object needs to define 'analysis_scheme'")
    scheme_name = c.analysis_scheme.lower()
    if scheme_name not in registry:
        raise NotImplementedError(f"Analysis scheme '{scheme_name}' is not available")

    module = importlib.import_module('NEDAS.schemes.'+scheme_name)
    Scheme = getattr(module, registry[scheme_name])

    return Scheme()



import unittest
import numpy as np
import re
from NEDAS.config.parse_config import convert_notation

class TestConvertNotation(unittest.TestCase):
    def test_dict_conversion(self):
        data = {"key1": "1e-3", "key2": "inf", "key3": "true"}
        expected = {"key1": 0.001, "key2": np.inf, "key3": True}
        self.assertEqual(convert_notation(data), expected)

    def test_list_conversion(self):
        data = ["-1e5", "-inf", "false"]
        expected = [-100000.0, -np.inf, False]
        self.assertEqual(convert_notation(data), expected)

    def test_nested_structure(self):
        data = {"outer": ["1e2", {"inner": "no"}]}
        expected = {"outer": [100.0, {"inner": False}]}
        self.assertEqual(convert_notation(data), expected)

    def test_string_to_bool(self):
        self.assertTrue(convert_notation("yes"))
        self.assertTrue(convert_notation("T"))
        self.assertFalse(convert_notation("no"))
        self.assertFalse(convert_notation(".false."))

    def test_string_to_float(self):
        self.assertEqual(convert_notation("1e-3"), 0.001)
        self.assertEqual(convert_notation("-1E4"), -10000.0)

    def test_inf_handling(self):
        self.assertEqual(convert_notation("inf"), np.inf)
        self.assertEqual(convert_notation("-inf"), -np.inf)

    def test_no_conversion(self):
        self.assertEqual(convert_notation("random string"), "random string")
        self.assertEqual(convert_notation(123), 123)
        self.assertEqual(convert_notation(["text", 45]), ["text", 45])

if __name__ == "__main__":
    unittest.main()



import typing
import numpy as np
from NEDAS.utils.netcdf_lib import nc_read_var, nc_write_var

def thickness_upper_limit(seaice_conc: np.ndarray, target_variable: typing.Literal['seaice','snow']) -> np.ndarray:
    """upper threshold for seaice_thick based on daily output of CICE 1993-2016"""
    upper_limit = np.full(seaice_conc.shape, 9999.)
    SIC0 = 0.55
    SIC1 = 0.8
    if target_variable == 'seaice':
        ind1 = np.where(np.logical_and(seaice_conc>SIC0, seaice_conc<SIC1))
        upper_limit[ind1] = 1.3 + np.exp(8.0*(seaice_conc[ind1]-0.76))
        ind2 = np.where(np.logical_and(seaice_conc>0, seaice_conc<SIC0))
        upper_limit[ind2] = 0.09 + 2.5*seaice_conc[ind2]

    elif target_variable == 'snow':
        ind1 = np.where(np.logical_and(seaice_conc>SIC0, seaice_conc<SIC1))
        upper_limit[ind1] = 0.2 + np.exp(2.0*(seaice_conc[ind1]-1.3))
        ind2 = np.where(np.logical_and(seaice_conc>0, seaice_conc<SIC0))
        upper_limit[ind2] = 0.03 + 0.7*seaice_conc[ind2]

    return upper_limit

def hi_cate(ncat, kcatb=0, kitd=1):
    """
    Defines ice thickness category boundaries based on the given scheme.
    
    Parameters:
        ncat (int): Number of ice thickness categories.
        kcatb (int): Determines the categorization scheme.
        kitd (int): Determines the remapping type (used when kcatb == 0).

    Returns:
        hilim (list of float): Ice thickness category boundaries.
    """
    hilim = np.zeros(ncat+1)  # Initialize the boundaries
    h_min = 0.01  # Default minimum ice thickness

    if kcatb == 0:
        if kitd == 1:  # Linear remapping
            cc1 = 3.0 / ncat
            cc2 = 15.0 * cc1
            cc3 = 3.0
            hilim[0] = 0.0
        else:  # Delta function remapping
            h_min = 0.1
            cc1 = max(1.1 / ncat, h_min)
            cc2 = 25.0 * cc1
            cc3 = 2.25
            hilim[0] = h_min

        for k in range(1, ncat + 1):
            x1 = (k - 2) / ncat
            hilim[k] = hilim[k - 1] + cc1 + cc2 * (1.0 + np.tanh(cc3 * (x1 - 1.0)))

    elif kcatb == 1:  # Linear scheme with additional coefficients
        cc1 = 3.0 / ncat
        cc2 = 0.5 / ncat
        hilim[0] = 0.0
        for k in range(1, ncat + 1):
            x1 = k - 1
            hilim[k] = x1 * (cc1 + (k - 2) * cc2)

    elif kcatb == 2:  # WMO standard categories
        if ncat == 5:
            hilim = [0.0, 0.3, 0.7, 1.2, 2.0, 999.0]
        elif ncat == 6:
            hilim = [0.0, 0.15, 0.3, 0.7, 1.2, 2.0, 999.0]
        elif ncat == 7:
            hilim = [0.0, 0.1, 0.15, 0.3, 0.7, 1.2, 2.0, 999.0]
        else:
            raise ValueError("kcatb=2 (WMO) must have ncat=5, 6, or 7")

    elif kcatb == -1:  # Single category
        hilim = [0.0, 100.0]

    else:  # Unsupported scheme
        raise ValueError("Error: no support for kcatb value in the current version!")

    return hilim

def fix_zsin_profile(Nlay, saltmax, depressT, nsal, msal):
    """
    Fixed salinity profile and melting temperature in ice layers
    only works under ktherm=1 for BL99 thermo
    """
    zSin = np.zeros(Nlay)
    Tmlt = np.zeros(Nlay)
    for k in range(Nlay-1):
        z = (k+0.5) / (Nlay-1)
        zSin[k] = 0.5*saltmax*(1-np.cos(np.pi*z**(nsal/(z+msal))))
        Tmlt[k] = -depressT*zSin[k]
    zSin[-1] = saltmax
    Tmlt[-1] = -depressT * saltmax
    return zSin, Tmlt

def adjust_ice_variables(prior_ice_file, post_ice_file,
                         fice, hice, mask,
                         aice_thresh, fice_thresh, hice_impact,
                         zSin, Tmlt) -> None:
    """
    Adjust the iced model restart file variables based on analysis of seaice properties: fice, hice
    """
    ncat = 5
    bkcat = hi_cate(ncat)

    aicen_f = nc_read_var(prior_ice_file, 'aicen')
    vicen_f = nc_read_var(prior_ice_file, 'vicen')
    vsnon_f = nc_read_var(prior_ice_file, 'vsnon')
    qsnon_f = nc_read_var(prior_ice_file, 'qsno001')

    ##bound aicen between 0,1
    aicen_f = np.maximum(np.minimum(aicen_f, 1), 0)

    aicen = aicen_f.copy()
    vicen = vicen_f.copy()
    vsnon = vsnon_f.copy()
    qicen_f = qsnon_f.copy()

    fice_f = np.sum(aicen_f, axis=0)
    aicen_f[np.where(np.logical_or(aicen_f <= aice_thresh, fice_f <= fice_thresh))] = 0.0
    fice_f = np.sum(aicen_f, axis=0)

    ##adjust for aicen, rescaled by analyzed fice
    ind = np.where(fice_f > fice_thresh)
    for k in range(ncat):
        aicen[k,...][ind] *= fice[ind] / fice_f[ind]
    ficem = np.sum(aicen, axis=0)
    ind = np.where(ficem > 1)
    for k in range(ncat):
        aicen[k,...][ind] /= ficem[ind]

    ind = np.where(fice_f <= fice_thresh)
    for k in range(ncat):
        aicen[k,...][ind] = 0.0
        vicen[k,...][ind] = 0.0
        vsnon[k,...][ind] = 0.0
    ficem = np.sum(aicen, axis=0)

    ind = np.where(np.logical_or(mask, fice <= fice_thresh))
    for k in range(ncat):
        aicen[k,...][ind] = 0.0
        vicen[k,...][ind] = 0.0
        vsnon[k,...][ind] = 0.0
        aicen_f[k,...][ind] = 0.0
        vicen_f[k,...][ind] = 0.0
    ficem[ind] = 0.0
    fice_f[ind] = 0.0

    ##adjust vicen
    ind = np.where(fice_f > fice_thresh)
    for k in range(ncat):
        ind1 = np.where(np.logical_and(aicen_f[k,...][ind] > 0, aicen[k,...][ind] > 0))
        vicen[k,...][ind][ind1] *= aicen[k,...][ind][ind1] / aicen_f[k,...][ind][ind1]
        #vsnon[k,...][ind][ind1] *= aicen[k,...][ind][ind1] / aicen_f[k,...][ind][ind1]

    ##when ice pack area assimilating hice
    ind1 = np.where(ficem[ind] > 0.75)
    sum_vice = np.sum(vicen, axis=0)[ind][ind1]
    Vtemp = ficem[ind][ind1] * (hice[ind][ind1]*hice_impact + sum_vice*(1-hice_impact)) / sum_vice
    for k in range(ncat):
        ind2 = np.where(aicen[k,...][ind][ind1] > 0)
        vicen[k,...][ind][ind1][ind2] *= Vtemp[ind2]
        ind2 = np.where(aicen[k,...][ind][ind1] <= 0)
        vicen[k,...][ind][ind1][ind2] = 0.0

    ##bound check again
    for k in range(ncat):
        ind2 = np.where(aicen[k,...][ind][ind1] > 0)
        vicen_min = aicen[k,...][ind][ind1][ind2] * bkcat[k]
        vicen_max = aicen[k,...][ind][ind1][ind2] * bkcat[k+1]
        vicen[k,...][ind][ind1][ind2] = np.maximum(np.minimum(vicen[k,...][ind][ind1][ind2], vicen_max), vicen_min)
        ind2 = np.where(aicen[k,...][ind][ind1] <= 0)
        vicen[k,...][ind][ind1][ind2] = 0.0
        vsnon[k,...][ind][ind1][ind2] = 0.0
        aicen[k,...][ind][ind1][ind2] = 0.0

    ind = np.where(fice_f <= fice_thresh)
    for k in range(ncat):
        vicen[k,...][ind] = 0.0
        vsnon[k,...][ind] = 0.0

    ind = np.where(np.logical_or(mask, ficem <= 0))
    for k in range(ncat):
        vicen[k,...][ind] = 0.0
        vsnon[k,...][ind] = 0.0
        aicen[k,...][ind] = 0.0
    ficem[ind] = 0.0

    ##replace the variables in posterior ice restart file
    ny, nx = fice.shape
    dims2 = {'nj':ny, 'ni':nx}
    dims3 = {'ncat':5, 'nj':ny, 'ni':nx}
    nc_write_var(post_ice_file, dims3, 'aicen', aicen, dtype=np.float64)
    nc_write_var(post_ice_file, dims3, 'vicen', vicen, dtype=np.float64)
    nc_write_var(post_ice_file, dims3, 'vsnon', vsnon, dtype=np.float64)

    ##other affected 2D variables
    ind = np.where(ficem <= 0)
    varlist_2d = ['uvel', 'vvel', 'strocnxT', 'strocnyT', 'frz_onset', 'iceumask']
    for i in range(4):
        varlist_2d.append(f'stressp_{i+1}')
        varlist_2d.append(f'stressm_{i+1}')
        varlist_2d.append(f'stress12_{i+1}')
    for vname in varlist_2d:
        var = nc_read_var(post_ice_file, vname)
        var[ind] = 0.0
        nc_write_var(post_ice_file, dims2, vname, var, dtype=np.float64)

    ##other affected 3D variables
    varlist_3d = ['iage', 'FY', 'alvl', 'vlvl', 'apnd', 'hpnd', 'ipnd', 'dhs', 'ffrac', 'Tsfcn']
    for i in range(7):
        varlist_3d.append(f'sice{i+1:03}')
        varlist_3d.append(f'qice{i+1:03}')
    varlist_3d.append('qsno001')
    for vname in varlist_3d:
        var = nc_read_var(post_ice_file, vname)
        for k in range(ncat):
            ind = np.where(np.logical_or(aicen[k,...] <= 0, ficem <= 0))
            if vname == 'Tsfcn':
                var[k,...][ind] = -1.8
                var[k,...][mask] = 0.0
            else:
                var[k,...][ind] = 0.0

            ind = np.where(aicen[k,...] > 0)
            if vname[:4] == 'sice':
                l = int(vname[-3:])
                var[k,...][ind] = zSin[l-1]

            #ind = np.where(aicen[k,...] <= aice_thresh)
            #if vname[:4] == 'qice':
            #    l = int(vname[-3:])
            #    Tmz = Tmlt[l-1]
            #    tmp1 = Tf[ind]+1.8
            #    Ti = -1.8 + tmp1*(l-0.5)/7
            #    var[k,...][ind] = -rhoi * (cp_ice*(Tmaz-Ti) + Lfresh*(1-Tmz/Ti) - cp_ocn*Tmz) / 7

        nc_write_var(post_ice_file, dims3, vname, var, dtype=np.float64)




import os
import inspect
from abc import ABC, abstractmethod
import numpy as np
from NEDAS.config import parse_config
from NEDAS.utils.progress import progress_bar

class Updator(ABC):
    """Base class for updators of the model restart files"""
    def __init__(self, c):
        ##get updator parameters from config file
        code_dir = os.path.dirname(inspect.getfile(self.__class__))
        config_dict = parse_config(code_dir, parse_args=False, **c.updator_def)
        for key, value in config_dict.items():
            setattr(self, key, value)

        self.increment = {}

    def update(self, c, state):
        """
        Top-level routine to apply the analysis increments to the original model
        restart files (as initial conditions for the next forecast)

        Inputs:
        - c: config module
        - state: state object, where field_prior, field_post are
        the field-complete state variables before and after assimilate()
        """
        pid_mem_show = [p for p,lst in state.mem_list.items() if len(lst)>0][0]
        pid_rec_show = [p for p,lst in state.rec_list.items() if len(lst)>0][0]
        c.pid_show = pid_rec_show * c.nproc_mem + pid_mem_show

        c.print_1p(f'>>> update model restart files with analysis increments\n')

        ##compute analysis increments
        self.compute_increment(c, state)

        ##process the fields, each processor goes through its own subset of
        ##mem_id,rec_id simultaneously
        ##but need to keep every rank in sync to coordinate multiprocess file access
        nm_max = np.max([len(lst) for _,lst in state.mem_list.items()])
        nr_max = np.max([len(lst) for _,lst in state.rec_list.items()])
        for r in range(nr_max):
            for m in range(nm_max):
                ##get file names for sync io
                pid_active = ( m < len(state.mem_list[c.pid_mem]) and r < len(state.rec_list[c.pid_rec]) )
                if pid_active:
                    mem_id = state.mem_list[c.pid_mem][m]
                    rec_id = state.rec_list[c.pid_rec][r]
                    rec = state.info['fields'][rec_id]
                    path = c.forecast_dir(rec['time'], rec['model_src'])
                    model = c.models[rec['model_src']]
                    file = model.filename(path=path, member=mem_id, **rec)
                else:
                    file = None
                all_files = c.comm.allgather(file)
                ##create the file locks
                for file in all_files:
                    c.comm.init_file_lock(file)

                if pid_active:
                    if c.debug:
                        print(f"PID {c.pid:4}: update_restartfile mem{mem_id+1:03} '{rec['name']:20}' {rec['time']} k={rec['k']}", flush=True)
                    else:
                        c.print_1p(progress_bar(m*nr_max+r, nm_max*nr_max))

                    ##apply the increment to restart files
                    self.update_restartfile(c, state, mem_id, rec_id)

        c.comm.Barrier()
        #c.comm.cleanup_file_locks()
        c.print_1p(' done.\n')

    @abstractmethod
    def compute_increment(self, c, state):
        pass

    @abstractmethod
    def update_restartfile(self, c, state, mem_id, rec_id):
        pass


"""Diagnostic module to plot the ensemble states"""

import os
import numpy as np
import matplotlib.pyplot as plt
from NEDAS.utils.conversion import ensure_list, t2h, h2t, dt1h
from NEDAS.utils.shell_utils import makedir
from NEDAS.utils.graphics import add_colorbar, adjust_ax_size, get_cmap

def get_task_list(c, **kwargs) -> list:

    variables = ensure_list(kwargs['variables'])
    model_src = ensure_list(kwargs['model_src'])
    vmin = ensure_list(kwargs['vmin'])
    vmax = ensure_list(kwargs['vmax'])
    nlevels = ensure_list(kwargs['nlevels'])
    cmap = ensure_list(kwargs['cmap'])

    tasks = []
    for member in range(c.nens):
        for i, vname in enumerate(variables):
            model = c.models[model_src[i]]
            levels = model.variables[vname]['levels']
            dt = model.variables[vname]['dt']
            for k in levels:
                for t in np.arange(t2h(c.time), t2h(c.next_time), dt):
                    tasks.append({**kwargs, 'time':h2t(t), 'member':member, 'model_src':model_src[i], 'vname':vname, 'k':k, 'vmin':vmin[i], 'vmax':vmax[i], 'nlevels':nlevels[i], 'cmap':cmap[i]})
    return tasks

def run(c, **kwargs) -> None:
    """
    Run diagnostics: plot the ensemble states
    """
    if 'plot_dir' in kwargs:
        plot_dir = kwargs['plot_dir']
    else:
        plot_dir = os.path.join(c.work_dir, 'plots', 'ensemble_states')
    makedir(plot_dir)

    figsize = (kwargs.get('fig_size_x', 9), kwargs.get('fig_size_y', 8))
    landcolor = kwargs.get('land_color', 'gray')

    variables = ensure_list(kwargs['variables'])
    vname = kwargs['vname']
    vmin = kwargs['vmin']
    vmax = kwargs['vmax']
    nlevels = kwargs['nlevels']
    cmap = get_cmap(kwargs['cmap'])

    member = kwargs['member']
    k = kwargs['k']
    time = kwargs['time']
    model_src = kwargs['model_src']

    if c.debug:
        print(f"PID {c.pid:4} plotting state variable '{vname:20}' k={k:3} at {time} for member{member+1:03}", flush=True)

    ##if the viewer html file does not exist, generate it
    viewer = os.path.join(plot_dir, 'index.html')
    if not os.path.exists(viewer):
        generate_viewer_html(c, plot_dir, model_src, variables, figsize)

    ##plot the variables defined in kwargs, save to figfile
    figfile = os.path.join(plot_dir, f"{vname}_k{k}_{time:%Y%m%dT%H%M%S}_mem{member+1:03}.png")

    ##read the field from model restart files
    model = c.models[model_src]
    if 'forecast_dir' in kwargs:
        fdir = kwargs['forecast_dir'].format(time=c.time)
    else:
        fdir = c.forecast_dir(c.time, model_src)
    model.read_grid(path=fdir, name=vname, k=k, member=member, time=time)
    var = model.read_var(path=fdir, name=vname, k=k, member=member, time=time)
    grid = model.grid

    rec = model.variables[vname]

    ##plot the field
    try:
        fig, ax = plt.subplots(1, 1, figsize=figsize)
        if rec['is_vector']:
            grid.plot_vectors(ax, var, V=vmax, showref=True, ref_units=rec['units'])
            adjust_ax_size(ax)
        else:
            grid.plot_field(ax, var, vmin=vmin, vmax=vmax, cmap=cmap)
            add_colorbar(fig, ax, cmap, vmin, vmax, nlevels, units=rec['units'])
        grid.plot_land(ax, color=landcolor)
        ax.set_title(f'member {member+1}', fontsize=16)
        ax.set_xlabel('x (m)', fontsize=14)
        ax.set_ylabel('y (m)', fontsize=14)
        fig.suptitle(f"{vname}, level {k:2}, {time}", fontsize=16)
        plt.savefig(figfile)
        plt.close()
    except Exception as e:
        print(f"ERROR: Failed to plot {vname} at level {k} and time {time} for member {member+1}")
        raise e

def generate_viewer_html(c, plot_dir, model_src, variables, figsize) -> None:
    """Generating a html page to help viewing the ensemble state variables"""
    if c.debug:
        print(f"Generating viewer.html page in {plot_dir}")

    with open(os.path.join(os.path.dirname(__file__), 'viewer.html'), 'rt') as f:
        html_page = f.read()

    ##replace the placeholder with the list of variables,levels,times,members
    levels_by_variable = ""
    times_by_variable = ""
    for vname in variables:
        levels_by_variable += f"'{vname}': ["
        model = c.models[model_src]
        for level in model.variables[vname]['levels']:
            levels_by_variable += f"{level}, "
        levels_by_variable += "], \n"
        times_by_variable += f"'{vname}': ["
        for t in np.arange(t2h(c.time_start), t2h(c.time_end), model.variables[vname]['dt']):
            times_by_variable += f"'{h2t(t):%Y%m%dT%H%M%S}', "
        times_by_variable += "], \n"
    html_page = html_page.replace("LEVELS_BY_VARIABLE", levels_by_variable)
    html_page = html_page.replace("TIMES_BY_VARIABLE", times_by_variable)

    members = "["
    for m in range(c.nens):
        members += f"'{m+1:03}', "
    members += "]"
    html_page = html_page.replace("MEMBERS", members)

    html_page = html_page.replace("TITLE", "Ensemble States")
    html_page = html_page.replace("IMAGE_WIDTH", f"{figsize[0]*60}")
    html_page = html_page.replace("IMAGE_HEIGHT", f"{figsize[1]*60}")

    ##write the html page to file
    with open(os.path.join(plot_dir, 'index.html'), 'w') as f:
        f.write(html_page)



import numpy as np

try:
    ###fft implementation using FFTW
    import pyfftw

    def fft2(f: np.ndarray) -> np.ndarray:
        """
        2D FFT implemented by pyFFTW. If pyFFTW is not available, will switch to np.fft.fft2

        Args:
            f (np.ndarray):
                Input field, of shape (..., ny, nx), of float32 type,
                the last two dimensions will be transformed by the 2D FFT.

        Returns:
            np.ndarray: Output field in spectral space, of complex64 type.
        """
        ##prepare fftw plan
        a = pyfftw.empty_aligned(f.shape, dtype='float32')
        b = pyfftw.empty_aligned(f.shape[:-1] + (f.shape[-1]//2+1,), dtype='complex64')
        fft_obj = pyfftw.FFTW(a, b, axes=(-2, -1))

        ##perform the fft2, output fh with same shape as f (keep dimension info)
        fh_ = fft_obj(f)
        fh = np.zeros(f.shape, dtype='complex64')

        nup = f.shape[-1]//2+1

        ##top half of the spectrum
        fh[..., 0:nup] = fh_

        ##the bottom half is conj of top half of the spectrum
        if f.shape[-1]%2 == 0:
            fh[..., 0, nup:] = np.conj(fh_[..., 0, nup-2:0:-1])
            fh[..., 1:, nup:] = np.conj(fh_[..., :0:-1, nup-2:0:-1])
        else:
            fh[..., 0, nup:] = np.conj(fh_[..., 0, nup-1:0:-1])
            fh[..., 1:, nup:] = np.conj(fh_[..., :0:-1, nup-1:0:-1])

        return fh

    def ifft2(fh: np.ndarray) -> np.ndarray:
        """
        Inverse 2D FFT implemented by pyFFTW. If pyFFTW is not available, will switch to np.fft.ifft2

        Args:
            fh (np.ndarray):
                Input field, of shape (..., nky, nkx), of complex64 type,
                the last two dimensions will be inverse transformed by 2D FFT.

        Returns:
            np.ndarray: Output field in physical space, of float32 type.
        """
        ##prepare fftw plan
        b = pyfftw.empty_aligned(fh.shape[:-1] + (fh.shape[-1]//2+1,), dtype='complex64')
        a = pyfftw.empty_aligned(fh.shape, dtype='float32')
        fft_obj = pyfftw.FFTW(b, a, axes=(-2, -1), direction='FFTW_BACKWARD')
        ##perform the ifft2
        ##TODO: check index
        ##TODO: normalize?
        f = fft_obj(fh[..., 0:fh.shape[1]//2+1])
        return f

except ImportError:
    #print("Warning: pyFFTW not found in your environment, will use numpy.fft instead.", flush=True)

    def fft2(x):
        return np.fft.fft2(x)

    def ifft2(x):
        return np.real(np.fft.ifft2(x))

def fftwn(n):
    """
    Wavenumber sequence corresponding to the FFT output in 1 dimension.

    Args:
        n (int): The size of the dimension

    Returns:
        np.ndarray: The sequence of wavenumber (0,1,2,...-2,-1) for this dimension
    """
    nup = int(np.ceil((n+1)/2))
    if n%2 == 0:
        wn = np.concatenate((np.arange(0, nup), np.arange(2-nup, 0)))
    else:
        wn = np.concatenate((np.arange(0, nup), np.arange(1-nup, 0)))
    return wn

def get_wn(fld):
    """
    Generates a meshgrid of wavenumbers corresponding to the input field dimensions.

    Args:
        fld (np.ndarray):
            Input field, the last two dimensions are the horizontal directions (ny, nx)

    Returns:
        np.ndarray:
            Wavenumbers corresponding to the input fields dimensions,
            relative to the domain size (nx or ny, whichever is larger).
    """
    ny, nx = fld.shape[-2:]

    wnx = np.zeros(fld.shape)
    wny = np.zeros(fld.shape)

    nup = int(max(nx, ny))

    for i in fftwn(nx):
        wnx[..., i] = i * nup / nx
    for j in fftwn(ny):
        wny[..., j, :] = j * nup / ny

    return wnx, wny




try:
    from numba import njit

except ImportError:
    print("Warning: numba is not found in your environment, will skip the njit precompiling of functions.", flush=True)

    ##define njit as en empty decorator
    from functools import wraps

    def njit(*njit_args, **njit_kwargs):
        #@njit() - decorator with arguments
        if len(njit_args) == 0:
            def decorator(func):
                @wraps(func)
                def wrapper(*func_args, **func_kwargs):
                    return func(*func_args, **func_kwargs)
                return wrapper
            return decorator

        #@njit - decorator without arguments
        elif callable(njit_args[0]):
            func = njit_args[0]
            @wraps(func)
            def wrapper(*func_args, **func_kwargs):
                return func(*func_args, **func_kwargs)
            return wrapper




import numpy as np
from NEDAS.utils.njit import njit
from NEDAS.assim_tools.assimilators.batch import BatchAssimilator

class ETKFAssimilator(BatchAssimilator):

    def local_analysis(self, c, loc_id, ind, hlfactor, state_data, obs_data):
        state_var_id = state_data['var_id']  ##variable id for each field (nfld)
        state_z = state_data['z'][:, loc_id]
        state_t = state_data['t'][:]

        ##vertical, time and cross-variable (impact_on_state) localization
        obs_value = obs_data['obs'][ind]
        obs_err = obs_data['err_std'][ind]
        obs_z = obs_data['z'][ind]
        obs_t = obs_data['t'][ind]
        obs_rec_id = obs_data['obs_rec_id'][ind]
        vroi = obs_data['vroi'][obs_rec_id]
        troi = obs_data['troi'][obs_rec_id]
        impact_on_state = obs_data['impact_on_state'][:, state_var_id][obs_rec_id]

        local_analysis_main(state_data['state_prior'][...,loc_id], obs_data['obs_prior'][:,ind],
                            obs_value, obs_err, hlfactor,
                            state_z, obs_z, vroi, c.localization_funcs['vertical'],
                            state_t, obs_t, troi, c.localization_funcs['temporal'],
                            impact_on_state, self.rfactor, self.kfactor, self.nlobs_max)

@njit
def local_analysis_main(state_prior, obs_prior,
                        obs, obs_err, hlfactor,
                        state_z, obs_z, vroi, vlocal_func,
                        state_t, obs_t, troi, tlocal_func,
                        impact_on_state, rfactor, kfactor, nlobs_max) -> None:
    """perform local analysis for one location in the analysis grid partition"""
    nens, nfld = state_prior.shape
    nens_obs, nlobs = obs_prior.shape
    if nens_obs != nens:
        raise ValueError('Error: number of ensemble members in state and obs do not match!')

    lfactor_old = np.zeros(nlobs)
    weights_old = np.eye(nens)

    ##loop through the field records
    for n in range(nfld):

        ##vertical localization
        vdist = np.abs(obs_z - state_z[n])
        vlfactor = vlocal_func(vdist, vroi)
        if (vlfactor==0).all():
            continue  ##the state is outside of vroi of all obs, skip

        ##temporal localization
        tdist = np.abs(obs_t - state_t[n])
        tlfactor = tlocal_func(tdist, troi)
        if (tlfactor==0).all():
            continue  ##the state is outside of troi of all obs, skip

        ##total lfactor
        lfactor =  hlfactor * vlfactor * tlfactor * impact_on_state[:, n]
        if (lfactor==0).all():
            continue

        ##if prior spread is zero, don't update
        if np.std(state_prior[:, n]) == 0:
            continue

        ##only need to assimilate obs with lfactor>0
        ind = np.where(lfactor>0)[0]

        ##TODO:get rid of obs if obs_prior is nan
        # valid = np.array([np.isnan(obs_prior[:,i]).any() for i in ind])
        # ind = ind[valid]

        ##sort the obs from high to low lfactor
        sort_ind = np.argsort(lfactor[ind])[::-1]
        ind = ind[sort_ind]

        ##limit number of local obs if needed
        ###e.g. topaz only keep the first 3000 obs with highest lfactor
        # nlobs_max = 3000
        ind = ind[:nlobs_max]

        ##use cached weight if no localization is applied, to avoid repeated computation
        if n>0 and len(ind)==len(lfactor_old) and (lfactor[ind]==lfactor_old).all():
            weights = weights_old
        else:
            weights = ensemble_transform_weights(obs[ind], obs_err[ind], obs_prior[:, ind], lfactor[ind], rfactor, kfactor)

        ##perform local analysis and update the ensemble state
        state_prior[:, n] = apply_ensemble_transform(state_prior[:, n], weights)

        lfactor_old = lfactor[ind]
        weights_old = weights

@njit
def ensemble_transform_weights(obs, obs_err, obs_prior, local_factor, rfactor, kfactor):
    nens, nlobs = obs_prior.shape

    ##ensemble weight matrix, weights[:, m] is for the m-th member
    ##also known as T in Bishop 2001, and X5 in Evensen textbook (and in Sakov 2012)
    weights = np.zeros((nens, nens))

    ##find mean of obs_prior
    obs_prior_mean = np.zeros(nlobs)
    for m in range(nens):
        obs_prior_mean += obs_prior[m, :]
    obs_prior_mean /= nens
    ##find variance of obs_prior
    obs_prior_var = np.zeros(nlobs)
    for m in range(nens):
        obs_prior_var += (obs_prior[m, :] - obs_prior_mean)**2
    obs_prior_var /= nens-1

    innov = obs - obs_prior_mean
    obs_var = obs_err**2

    ##inflate obs error by rfactor, and kfactor where innovation is large
    obs_var = np.sqrt((obs_prior_var + obs_var)**2 + obs_prior_var*(innov/kfactor)**2) - obs_prior_var
    obs_err = np.sqrt(obs_var)

    ##obs_prior_pert S and innovation dy, normalized by sqrt(nens-1) R^-0.5
    S = np.zeros((nlobs, nens))
    dy = np.zeros((nlobs))
    for p in range(nlobs):
        S[p, :] = (obs_prior[:, p] - obs_prior_mean[p]) * local_factor[p] / obs_err[p]
        dy[p] = (obs[p] - obs_prior_mean[p]) * local_factor[p] / obs_err[p]
    S /= np.sqrt(nens-1)
    dy /= np.sqrt(nens-1)

    ##TODO:factor in the correlated R in obs_err, need another SVD of R
    # obs_err_corr

    ##----first part of weights: update of mean
    ##find singular values of the inverse variance ratio matrix (I + S^T S)
    ##note: the added I actually helps prevent issues if S^T S is not full rank
    ##      when nlobs<nens, there will be singular values of 0, but the full matrix
    ##      can still be inverted with singular values of 1.
    var_ratio_inv = np.eye(nens) + S.T @ S

    ##TODO: var_ratio_inv = np.eye(nlobs) + S @ S.T  if nlobs<nens

    try:
        L, sv, Rh = np.linalg.svd(var_ratio_inv)
    except:
        ##if svd failed just return equal weights (no update)
        print('Error: failed to invert var_ratio_inv=', var_ratio_inv)
        return np.eye(nens)

    ##the update of ens mean is given by (I + S^T S)^-1 S^T dy
    ##namely, var_ratio * obs_prior_var / obs_var * dy = G dy
    var_ratio = L @ np.diag(sv**-1) @ Rh

    ##the gain matrix
    gain = var_ratio @ S.T

    for m in range(nens):
        weights[m, :] = np.sum(gain[m, :] * dy)

    ##---second part of weights: update of ensemble spread
    if rfactor > 1:  ##scale S with rfactor and recompute the var_ratio
        S /= np.sqrt(rfactor)
        try:
            var_ratio_inv = np.eye(nens) + S.T @ S
            L, sv, Rh = np.linalg.svd(var_ratio_inv)
            var_ratio = L @ np.diag(sv**-1) @ Rh
            gain = var_ratio @ S.T
        except:
            ##if failed just return equal weights (no update)
            return np.eye(nens)

    var_ratio_sqrt = L @ np.diag(sv**-0.5) @ Rh

    weights += var_ratio_sqrt

    return weights

@njit
def apply_ensemble_transform(ens_prior, weights):
    """Apply the weights to transform local ensemble"""

    nens = ens_prior.size
    ens_post = ens_prior.copy()

    ##check if weights sum to 1
    for m in range(nens):
        sum_wgts = np.sum(weights[:, m])
        if np.abs(sum_wgts - 1) > 1e-5:
            print('Warning: sum of weights != 1 detected!')

    ##apply the weights
    for m in range(nens):
        ens_post[m] = np.sum(ens_prior * weights[:, m])

    return ens_post




import os
import numpy as np
from NEDAS.assim_tools.updators.base import Updator

class AlignmentUpdator(Updator):
    """Updator class with alignment technique"""

    def compute_increment(self, c, state):
        """
        Alignment technique: compute optical flows from the pair of prior/posterior state variable field
        """
        c.print_1p("Compute alignment based on analysis increment of '"+c.alignment['variable']+"'...\n")

        for rec_id in state.rec_list[c.pid_rec]:
            rec = state.info['fields'][rec_id]
            model = c.models[rec['model_src']]
            path = c.forecast_dir(rec['time'], rec['model_src'])
            if rec['name'] != c.alignment['variable']:
                continue

            for mem_id in state.mem_list[c.pid_mem]:
                ##compute displacement on analysis grid
                fld_prior = state.fields_prior[mem_id, rec_id]
                fld_post = state.fields_post[mem_id, rec_id]
                displace = optical_flow(c.grid, fld_prior, fld_post, **c.alignment)
                np.save(os.path.join(state.analysis_dir, f"displace.m{mem_id}.k{rec['k']}.npy"), displace)

                ##the model class offer a method to update grid (a lagrangian approach)
                ##so displace increment will be applied directly to the grid elements
                if hasattr(model, 'displace'):
                    ##convert the displacement from analysis grid to model grid
                    model.read_grid(path=path, member=mem_id, **rec)
                    c.grid.set_destination_grid(model.grid)
                    displace_m = c.grid.convert(displace, is_vector=True, method='linear')
                    ##apply the displacement
                    model.displace(displace_m[0,...], displace_m[1,...], path=path, member=mem_id, **rec)

        c.comm.Barrier()

    def update_restartfile(self, c, state, mem_id, rec_id):
        """
        Alignment technique, use the displace increment to adjust the model grid to
        precondition all the analysis variables for next assimilation step
        See more details in Ying 2019
        """
        rec = state.info['fields'][rec_id]
        model = c.models[rec['model_src']]
        path = c.forecast_dir(rec['time'], rec['model_src'])
        fld_prior = state.fields_prior[mem_id, rec_id]
        fld_post = state.fields_post[mem_id, rec_id]

        ##get model state variable prior
        model.read_grid(path=path, member=mem_id, **rec)
        c.grid.set_destination_grid(model.grid)
        var_prior = model.read_var(path=path, member=mem_id, **rec)

        if rec['is_vector']:
            fld_shape = var_prior.shape[1:]
        else:
            fld_shape = var_prior.shape

        ##read the corresponding displacement and convert to model grid
        displace = np.load(os.path.join(state.analysis_dir, f"displace.m{mem_id}.k{rec['k']}.npy"))

        ##warp the prior field with displacement
        fld_prior_warp = fld_prior.copy()
        u = displace[0,...]/c.grid.dx
        v = displace[1,...]/c.grid.dx
        for ind in np.ndindex(fld_prior.shape[:-2]):
            fld_prior_warp[ind] = warp(fld_prior[ind], -u, -v)

        ##residual increments not explained by the displacement
        res_incr = fld_post - fld_prior_warp

        if hasattr(model, 'displace'):
            ##apply the residual increment
            res_incr_m = c.grid.convert(res_incr, is_vector=rec['is_vector'], method='linear')
            if fld_shape == model.grid.x.shape:
                var_post = var_prior + res_incr_m
            elif fld_shape == model.grid.x_elem.shape:
                var_post = var_prior + np.mean(res_incr_m[...,model.grid.tri.triangles], axis=-1)
            else:
                raise RuntimeError(f"mismatch in field prior {var_prior.shape} with residual increment {res_incr_m.shape}")

        else:
            new_var = c.grid.convert(fld_post, is_vector=rec['is_vector'], method='linear')
            if fld_shape == model.grid.x.shape:
                var_post = new_var
            elif fld_shape == model.grid.x_elem.shape:
                var_post = np.mean(new_var[...,model.grid.tri.triangles], axis=-1)
            else:
                raise RuntimeError(f"mismatch in field prior {var_prior.shape} with posterior {new_var.shape}")

        #write the posterior variable to restart file
        ind = np.where(np.isnan(var_post))
        var_post[ind] = var_prior[ind]
        #if np.isnan(var_post).any():
        #    raise ValueError('nan detected in var_post')
        model.write_var(var_post, path=path, member=mem_id, comm=c.comm, **rec)

def optical_flow(grid, fld1, fld2, nlevel=5, niter_max=100, smoothness_weight=1, **kwargs):
    ni = int(2**np.ceil(np.log(np.max(fld1.shape))/np.log(2)))
    x1 = np.full((ni,ni), np.nan)
    x2 = np.full((ni,ni), np.nan)
    x1[0:grid.ny, 0:grid.nx] = fld1.copy()
    x2[0:grid.ny, 0:grid.nx] = fld2.copy()
    mask = np.logical_or(np.isnan(x1), (np.abs(x2-x1)<0.00001))
    x1[mask] = 0
    x2[mask] = 0
    w = smoothness_weight
    ni, nj = x1.shape
    ##normalize field so that w can be fixed
    xmax = np.max(x1[:, :]); xmin = np.min(x1[:, :])
    if (xmax>xmin):
        x1[:, :] = (x1[:, :] - xmin) / (xmax -xmin)
        x2[:, :] = (x2[:, :] - xmin) / (xmax -xmin)
    u = np.zeros((ni, nj))
    v = np.zeros((ni, nj))
    ###pyramid levels
    for lev in range(nlevel, -1, -1):
        x1w = warp(x1, -u, -v)
        x1c = coarsen(x1w, 1, lev)
        x2c = coarsen(x2, 1, lev)
        maskc = coarsen_mask(mask, 1, lev)
        xdx = 0.5*(deriv_x(x1c) + deriv_x(x2c))
        xdy = 0.5*(deriv_y(x1c) + deriv_y(x2c))
        xdt = x2c - x1c
        ###compute incremental flow using iterative solver
        du = np.zeros(xdx.shape)
        dv = np.zeros(xdx.shape)
        du1 = np.zeros(xdx.shape)
        dv1 = np.zeros(xdx.shape)
        niter = 0
        diff = 1e7
        while diff > 1e-3 and niter < niter_max:
            du[0,:] = 0; du[-1,:] = 0; du[:,0] = 0; du[:,-1] = 0  ##boundary conditions
            dv[0,:] = 0; dv[-1,:] = 0; dv[:,0] = 0; dv[:,-1] = 0
            du[maskc] = 0
            dv[maskc] = 0
            ubar = laplacian(du) + du
            vbar = laplacian(dv) + dv
            du1 = ubar - xdx*(xdx*ubar + xdy*vbar + xdt)/(w + xdx**2 + xdy**2)
            dv1 = vbar - xdy*(xdx*ubar + xdy*vbar + xdt)/(w + xdx**2 + xdy**2)
            diff = np.max(np.abs(du1-du) + np.abs(dv1-dv))
            du = du1
            dv = dv1
            niter += 1
        #print(niter, diff)
        u += sharpen(du*2**(lev-1), lev, 1)
        v += sharpen(dv*2**(lev-1), lev, 1)

    disp_u = u[0:grid.ny, 0:grid.nx] * grid.dx
    disp_v = v[0:grid.ny, 0:grid.nx] * grid.dy
    return np.array([disp_u, disp_v])

def warp(x, u, v):
    xw = x.copy()
    ni, nj = x.shape
    ii, jj = np.mgrid[0:ni, 0:nj]
    xw = interp2d(x, ii+v, jj+u)
    return xw

def coarsen_mask(x, lev1, lev2):  ##only subsample no smoothing, avoid mask growing
    if lev1 < lev2:
        for _ in range(lev1, lev2):
            ni, nj = x.shape
            x1 = x[0:ni:2, 0:nj:2]
            x = x1
    return x

def coarsen(x, lev1, lev2):
    if lev1 < lev2:
        for _ in range(lev1, lev2):
            ni, nj = x.shape
            x1 = 0.25*(x[0:ni:2, 0:nj:2] + x[1:ni:2, 0:nj:2] + x[0:ni:2, 1:nj:2] + x[1:ni:2, 1:nj:2])
            x = x1
    return x

def sharpen(x, lev1, lev2):
    if lev1 > lev2:
        for _ in range(lev1, lev2, -1):
            ni, nj = x.shape
            x1 = np.zeros((ni*2, nj))
            x1[0:ni*2:2, :] = x
            x1[1:ni*2:2, :] = 0.5*(np.roll(x, -1, axis=0) + x)
            x2 = np.zeros((ni*2, nj*2))
            x2[:, 0:nj*2:2] = x1
            x2[:, 1:nj*2:2] = 0.5*(np.roll(x1, -1, axis=1) + x1)
            x = x2
    return x

def interp2d(x, io, jo):
    ni, nj = x.shape
    io1 = np.floor(io).astype(int) % ni
    jo1 = np.floor(jo).astype(int) % nj
    io2 = np.floor(io+1).astype(int) % ni
    jo2 = np.floor(jo+1).astype(int) % nj
    di = io - np.floor(io)
    dj = jo - np.floor(jo)
    xo = (1-di)*(1-dj)*x[io1, jo1] + di*(1-dj)*x[io2, jo1] + (1-di)*dj*x[io1, jo2] + di*dj*x[io2, jo2]
    return xo

def deriv_y(f):
    return 0.5*(np.roll(f, -1, axis=0) - np.roll(f, 1, axis=0))

def deriv_x(f):
    return 0.5*(np.roll(f, -1, axis=1) - np.roll(f, 1, axis=1))

def laplacian(f):
    out = (np.roll(f, -1, axis=1) + np.roll(f, 1, axis=1) + np.roll(f, -1, axis=0) + np.roll(f, 1, axis=0))/6
    out += (np.roll(np.roll(f, -1, axis=0), -1, axis=1) + np.roll(np.roll(f, -1, axis=0), 1, axis=1) + np.roll(np.roll(f, 1, axis=0), -1, axis=1) + np.roll(np.roll(f, 1, axis=0), 1, axis=1))/12
    out -= f
    return out



import os
import netCDF4
import numpy as np

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import tensorflow as tf

class Att_Res_UNet():
    def __init__(self, list_predictors, list_targets, patch_dim, batch_size, n_filters, activation, kernel_initializer, batch_norm, pooling_type, dropout):
        self.list_predictors = list_predictors
        self.list_targets = list_targets
        self.patch_dim = patch_dim
        self.batch_size = batch_size
        self.n_filters = n_filters
        self.activation = activation
        self.kernel_initializer = kernel_initializer
        self.batch_norm = batch_norm
        self.pooling_type = pooling_type
        self.dropout = dropout
        self.n_predictors = len(list_predictors)
        self.n_targets = len(list_targets)

    def repeat_elem(self, tensor, rep):
        return tf.keras.layers.Lambda(lambda x, repnum: tf.keras.backend.repeat_elements(x, repnum, axis = 3), arguments = {'repnum': rep})(tensor)

    def gating_signal(self, x, n_filters, batch_norm = False):
        x = tf.keras.layers.Conv2D(n_filters, (1,1), padding = "same")(x)
        if batch_norm == True:
            x = tf.keras.layers.BatchNormalization()(x)
        x = tf.keras.layers.Activation("relu")(x)
        return(x)

    def attention_block(self, x, g, inter_shape):
        shape_x = tf.keras.backend.int_shape(x)
        shape_g = tf.keras.backend.int_shape(g)

        theta_x = tf.keras.layers.Conv2D(inter_shape, kernel_size = (2,2), strides = (2,2), padding = "same")(x)
        shape_theta_x = tf.keras.backend.int_shape(theta_x)

        phi_g = tf.keras.layers.Conv2D(inter_shape, kernel_size = (1,1), padding = "same")(g)
        upsample_g = tf.keras.layers.Conv2DTranspose(inter_shape, (3,3), 
                                                     strides = (shape_theta_x[1] // shape_g[1], shape_theta_x[2] // shape_g[2]),
                                                     padding = "same")(phi_g)

        concat_xg = tf.keras.layers.add([upsample_g, theta_x])
        act_xg = tf.keras.layers.Activation("relu")(concat_xg)

        psi = tf.keras.layers.Conv2D(1, (1,1), padding = "same")(act_xg)
        sigmoid_xg = tf.keras.layers.Activation("sigmoid")(psi)
        shape_sigmoid = tf.keras.backend.int_shape(sigmoid_xg)

        upsample_psi = tf.keras.layers.UpSampling2D(size = (shape_x[1] // shape_sigmoid[1], shape_x[2] // shape_sigmoid[2]))(sigmoid_xg)
        upsample_psi = self.repeat_elem(upsample_psi, shape_x[3])
        y = tf.keras.layers.multiply([upsample_psi, x])

        result = tf.keras.layers.Conv2D(shape_x[3], (1,1), padding = "same")(y)
        result_bn = tf.keras.layers.BatchNormalization()(result)

        return(result_bn)

    def residual_conv_block(self, x, n_filters, padding = "same", kernel_size=(3,3)):
        conv = tf.keras.layers.Conv2D(n_filters, kernel_size = kernel_size, padding = padding, kernel_initializer = self.kernel_initializer)(x)
        if self.batch_norm == True:
            conv = tf.keras.layers.BatchNormalization(axis = 3)(conv)
        conv = tf.keras.layers.Activation(self.activation)(conv)

        conv = tf.keras.layers.Conv2D(n_filters, kernel_size = kernel_size, padding = padding, kernel_initializer = self.kernel_initializer)(conv)
        if self.batch_norm == True:
            conv = tf.keras.layers.BatchNormalization(axis = 3)(conv)

        shortcut = tf.keras.layers.Conv2D(n_filters, kernel_size = (1,1), padding = padding)(x)
        if self.batch_norm == True:
            shortcut = tf.keras.layers.BatchNormalization(axis = 3)(shortcut)

        res_path = tf.keras.layers.add([shortcut, conv])
        res_path = tf.keras.layers.Activation(self.activation)(res_path)

        return(res_path)

    def downsample_block(self, x, n_filters, pool_size = (2,2), kernel_size = (3,3), strides = 2):
        f = self.residual_conv_block(x, n_filters, kernel_size=kernel_size)

        if self.pooling_type == "Max":
            p = tf.keras.layers.MaxPool2D(pool_size = pool_size, strides = strides)(f)
        elif self.pooling_type == "Average":
            p = tf.keras.layers.AveragePooling2D(pool_size = pool_size, strides = strides)(f)

        p = tf.keras.layers.Dropout(self.dropout)(p)
        return(f, p)  

    def upsample_block(self, x, conv_features, n_filters, kernel_size = (3,3), strides = 2, padding = "same"):
        gating = self.gating_signal(x, n_filters)
        att = self.attention_block(conv_features, gating, n_filters)
        up_att = tf.keras.layers.UpSampling2D(size = (2, 2), data_format = "channels_last")(x)
        up_att = tf.keras.layers.concatenate([up_att, att], axis = 3)
        up_conv = self.residual_conv_block(up_att, n_filters,kernel_size=kernel_size)
        return(up_conv)

    def make_unet_model(self): 
        inputs = tf.keras.layers.Input(shape = (*self.patch_dim, self.n_predictors))
        # Encoder (downsample)
        f1, p1 = self.downsample_block(inputs, self.n_filters[0], kernel_size=(7,7))
        f2, p2 = self.downsample_block(p1, self.n_filters[1])
        f3, p3 = self.downsample_block(p2, self.n_filters[2])
        f4, p4 = self.downsample_block(p3, self.n_filters[3])
        f5, p5 = self.downsample_block(p4, self.n_filters[4])
        # Bottleneck
        u5 = self.residual_conv_block(p5, self.n_filters[5])
        # Decoder (upsample)
        u4 = self.upsample_block(u5, f5, self.n_filters[4])
        u3 = self.upsample_block(u4, f4, self.n_filters[3])
        u2 = self.upsample_block(u3, f3, self.n_filters[2])
        u1 = self.upsample_block(u2, f2, self.n_filters[1])
        u0 = self.upsample_block(u1, f1, self.n_filters[0])
        # outputs
        SICerror = tf.keras.layers.Conv2D(self.n_targets, (1, 1), padding = "same", activation = "linear", dtype = tf.float32, name = "psi")(u0)
        unet_model = tf.keras.Model(inputs, SICerror, name = "U-Net")

        return(unet_model)

    def featname2tuple(feature_name):
        parts = feature_name.rsplit('_', 1)  # Split from the right, max 1 split
        varname = parts[0]
        channel = int(parts[1])
        return varname, channel

class Data_generator(tf.keras.utils.Sequence):
    def __init__(self, nrun, startrun, shuffle, batch_size, dim,  path_data, list_predictors, list_targets, sampleperrun = 100):
        self.nrun = nrun
        self.startrun = startrun
        self.sampleperrun = sampleperrun
        self.n = nrun*(sampleperrun-1)
        self.shuffle = shuffle
        self.batch_size = batch_size
        self.path_data = path_data
        self.dim = dim
        self.shuffle = shuffle
        self.list_predictors = list_predictors
        self.list_targets = list_targets
        self.npredictors = len(self.list_predictors)
        self.ntargets = len(self.list_targets)
        self.indexes = np.arange(self.n)
        if self.shuffle == True:
            rng = np.random.default_rng()
            rng.shuffle(self.indexes)

    def __len__(self): # Number of batches per epoch
        return self.n // self.batch_size

    def index2rs(self, index): #From index of the sample to the number of the run and sample
        r = index // self.sampleperrun
        s = index % self.sampleperrun
        return r,s

    def __getitem__(self, index): # Generate one batch of data
        indexes = self.indexes[index * self.batch_size : (index + 1) * self.batch_size]
        list_r_s = [self.index2rs(k) for k in indexes]
        X, y = self.__data_generation(list_r_s)
        return (X, y)

    def on_epoch_end(self): # Updates indexes after each epoch
        self.indexes = np.arange(self.n)
        if self.shuffle == True:
            rng = np.random.default_rng()
            rng.shuffle(self.indexes)
    def __data_generation(self, list_r_s):  # Generates data containing batch_size samples

        X = np.full((self.batch_size, *self.dim, self.npredictors), np.nan)
        y = np.full((self.batch_size, *self.dim, self.ntargets), np.nan)

        for i, (r,s) in enumerate(list_r_s):
            fileIDX = os.path.join(self.path_data,f'{r+1+self.startrun:04d}',f'{s:03d}.nc')
            fileIDy = os.path.join(self.path_data,f'{r+1+self.startrun:04d}',f'{s+1:03d}.nc')
            ncx = netCDF4.Dataset(fileIDX, "r")
            ncy = netCDF4.Dataset(fileIDy, "r")
            for k in range(self.npredictors):
                varname, channel = featname2tuple(self.list_predictors[k])
                X[i,...,k] = ncx.variables[varname][0,channel,:,:]
            for k in range(self.ntargets):
                varname, channel = featname2tuple(self.list_targets[k])
                y[i,...,k] = ncy.variables[varname][0,channel,:,:]
            ncx.close()
            ncy.close()
        return (X, y)




import copy
import numpy as np
from matplotlib import colormaps
from NEDAS.grid.grid_base import GridBase

class RegularGrid(GridBase):
    """
    Regular 2D grid class

    Args:
        pole_dim (str, optional):
            `None` (default), if one of the dimension has poles, :code:`'x'` or :code:`'y'`
        pole_index (str, optional):
            `None` (default), tuple of the pole index(s) in `pole_dim`
        neighbors (np.ndarray, optional):
            `None` (default), for regular grid with special geometry (e.g. tripolar ocean grid),
            `neighbors` stores the j,i index of 4 neighors (east, north, west, and south) on each grid point.
            Since specifying neighbors already take care of cyclic boundary conditions, `cyclic_dim` will be discarded if `neighbors` is set.

    """
    def __init__(self, proj, x, y, bounds=None, cyclic_dim=None, distance_type='cartesian',
                 pole_dim=None, pole_index=None, neighbors=None, dst_grid=None,):
        super().__init__(proj, x, y, bounds, cyclic_dim, distance_type, dst_grid)
        self.regular = True
        self.pole_dim = pole_dim
        self.pole_index = pole_index
        self.neighbors = neighbors
        if self.neighbors is not None and self.cyclic_dim is not None:
            print('neighbors already implemented, discarding cyclic_dim')
            self.cyclic_dim = None

        self.nx = self.x.shape[1]
        self.ny = self.x.shape[0]
        self.dx = (self.xmax - self.xmin) / (self.nx - 1)
        self.dy = (self.ymax - self.ymin) / (self.ny - 1)
        self.Lx = self.nx * self.dx
        self.Ly = self.ny * self.dy
        self.npoints = self.nx * self.ny

    def change_resolution_level(self, nlevel):
        """
        Generate a new grid with changed resolution.

        Args:
            nlevel (int):
                Positive number, downsample grid abs(nlevel) times, each time doubling the grid spacing;
                Negative number, upsample grid abs(nlevel) times, each time halving the grid spacing

        Returns:
            A new grid object with changed resolution.
        """
        if nlevel == 0:
            return self
        else:
            ##create a new grid object with x,y at new resolution level
            self._dst_grid = None
            new_grid = copy.deepcopy(self)
            fac = 2**nlevel
            new_grid.dx = self.dx * fac
            new_grid.dy = self.dy * fac
            new_grid.nx = int(np.round(self.Lx / new_grid.dx))
            new_grid.ny = int(np.round(self.Ly / new_grid.dy))
            assert min(new_grid.nx, new_grid.ny) > 1, "Grid.change_resolution_level: new resolution too low, try smaller nlevel"
            new_grid.x, new_grid.y = np.meshgrid(self.xmin + np.arange(new_grid.nx) * new_grid.dx, self.ymin + np.arange(new_grid.ny) * new_grid.dy)
            ##coarsen the mask
            self.set_destination_grid(new_grid)
            new_grid.mask = self.convert(self.mask, method='nearest').astype(bool)
            return new_grid

    def find_index(self, x_, y_):
        x_ = np.array(x_).flatten()
        y_ = np.array(y_).flatten()

        ##lon: pyproj.Proj works only for lon=-180:180
        if self.proj_name == 'longlat':
            x_ = np.mod(x_ + 180., 360.) - 180.

        ###account for cyclic dim, when points drop "outside" then wrap around
        x_, y_ = self._wrap_cyclic_xy(x_, y_)

        xi = self.x[0, :]
        yi = self.y[:, 0]

        ##sort the index to monoticially increasing
        ##x_,y_ are the sorted coordinates of grid points
        ##i_,j_ are their original grid index
        i_ = np.argsort(xi)
        xi_ = xi[i_]
        j_ = np.argsort(yi)
        yi_ = yi[j_]

        ##pad cyclic dimensions with additional grid point for the wrap-around
        if self.cyclic_dim is not None:
            for d in self.cyclic_dim:
                if d=='x':
                    if xi_[0]+self.Lx not in xi_:
                        xi_ = np.hstack((xi_, xi_[0] + self.Lx))
                        i_ = np.hstack((i_, i_[0]))
                elif d=='y':
                    if yi_[0]+self.Ly not in yi_:
                        yi_ = np.hstack((yi_, yi_[0] + self.Ly))
                        j_ = np.hstack((j_, j_[0]))

        ##if neighbors indices are provided, the search range is extended by 1 grid on both sides
        if self.neighbors is not None:
            xi_ = np.hstack((xi_[0]-self.dx, xi_, xi_[-1]+self.dx))
            i_ = np.hstack((i_[0]-1, i_, i_[-1]+1))
            yi_ = np.hstack((yi_[0]-self.dy, yi_, yi_[-1]+self.dy))
            j_ = np.hstack((j_[0]-1, j_, j_[-1]+1))

        ##now find the position near the given x_,y_ coordinates
        ##pi,pj are the index in the padded array, right side of the given x_,y_
        ##only the positions inside the grid will be kept
        pi = np.array(np.searchsorted(xi_, x_, side='right'))
        pj = np.array(np.searchsorted(yi_, y_, side='right'))
        inside = ~np.logical_or(np.logical_or(pi==len(xi_), pi==0),
                                np.logical_or(pj==len(yi_), pj==0))
        pi, pj = pi[inside], pj[inside]

        ##vertices (p1, p2, p3, p4) for the rectangular grid box
        ##p3 is the point found by the search index (pj,pi),
        ##internal coordinates (in_x, in_y) pinpoint the x_,y_ location inside
        ##the rectangle with values range [0, 1)
        ##(pj,pi-1)   p4----+------p3 (pj,pi)
        ##             |    |      |
        ##             +in_x*------+
        ##             |    in_y   |
        ##(pj-1,pi-1) p1----+------p2 (pj-1,pi)
        indices = None #for regular grid, the element indices are not used

        if self.neighbors is not None:
            ##find the right indices for each vertex grid point
            j1,i1 = j_[pj-1], i_[pi-1]
            j2,i2 = np.zeros(pj.shape, dtype=int), np.zeros(pj.shape, dtype=int)
            j3,i3 = np.zeros(pj.shape, dtype=int), np.zeros(pj.shape, dtype=int)
            j4,i4 = np.zeros(pj.shape, dtype=int), np.zeros(pj.shape, dtype=int)

            ind = np.where(np.logical_and(j1>=0, i1>=0)) ##p1 is the anchor in neighbors
            j2[ind], i2[ind] = self.neighbors[0,0,j1[ind],i1[ind]], self.neighbors[1,0,j1[ind],i1[ind]]
            j3[ind], i3[ind] = self.neighbors[0,1,j2[ind],i2[ind]], self.neighbors[1,1,j2[ind],i2[ind]]
            j4[ind], i4[ind] = self.neighbors[0,1,j1[ind],i1[ind]], self.neighbors[1,1,j1[ind],i1[ind]]

            ind = np.where(np.logical_and(j1>=0, i1<0)) ##p2 is the anchor in neighbors
            j2[ind], i2[ind] = j_[pj-1][ind], i_[pi][ind]
            j1[ind], i1[ind] = self.neighbors[0,2,j2[ind],i2[ind]], self.neighbors[1,2,j2[ind],i2[ind]]
            j3[ind], i3[ind] = self.neighbors[0,1,j2[ind],i2[ind]], self.neighbors[1,1,j2[ind],i2[ind]]
            j4[ind], i4[ind] = self.neighbors[0,2,j3[ind],i3[ind]], self.neighbors[1,2,j3[ind],i3[ind]]

            ind = np.where(np.logical_and(j1<0, i1<0)) ##p3 is the anchor in neighbors
            j3[ind], i3[ind] = j_[pj][ind], i_[pi][ind]
            j2[ind], i2[ind] = self.neighbors[0,3,j3[ind],i3[ind]], self.neighbors[1,3,j3[ind],i3[ind]]
            j4[ind], i4[ind] = self.neighbors[0,2,j3[ind],i3[ind]], self.neighbors[1,2,j3[ind],i3[ind]]
            j1[ind], i1[ind] = self.neighbors[0,2,j2[ind],i2[ind]], self.neighbors[1,2,j2[ind],i2[ind]]

            ind = np.where(np.logical_and(j1<0, i1>=0)) ##p4 is the anchor in neighbors
            j4[ind], i4[ind] = j_[pj][ind], i_[pi-1][ind]
            j1[ind], i1[ind] = self.neighbors[0,3,j4[ind],i4[ind]], self.neighbors[1,3,j4[ind],i4[ind]]
            j3[ind], i3[ind] = self.neighbors[0,0,j4[ind],i4[ind]], self.neighbors[1,0,j4[ind],i4[ind]]
            j2[ind], i2[ind] = self.neighbors[0,0,j1[ind],i1[ind]], self.neighbors[1,0,j1[ind],i1[ind]]

        else:
            ##use normal rectangle grid indices
            j1, i1 = j_[pj-1], i_[pi-1]
            j2, i2 = j_[pj-1], i_[pi]
            j3, i3 = j_[pj],   i_[pi]
            j4, i4 = j_[pj],   i_[pi-1]

        ##assign the points to vertices
        vertices = np.zeros(pi.shape+(4,), dtype=int)
        vertices[:, 0] = j1 * self.nx + i1
        vertices[:, 1] = j2 * self.nx + i2
        vertices[:, 2] = j3 * self.nx + i3
        vertices[:, 3] = j4 * self.nx + i4

        ##internal coordinates inside rectangles
        in_coords = np.zeros(pi.shape+(2,), dtype=np.float64)
        in_coords[:, 0] = (x_[inside] - xi_[pi-1]) / (xi_[pi] - xi_[pi-1])
        in_coords[:, 1] = (y_[inside] - yi_[pj-1]) / (yi_[pj] - yi_[pj-1])

        ##index of grid point nearest to (x_,y_)
        j_near = np.zeros(pj.shape, dtype=int)
        i_near = np.zeros(pj.shape, dtype=int)
        ind = np.where(np.logical_and(in_coords[:,0]<0.5, in_coords[:,1]<0.5))
        j_near[ind], i_near[ind] = j1[ind], i1[ind]
        ind = np.where(np.logical_and(in_coords[:,0]>=0.5, in_coords[:,1]<0.5))
        j_near[ind], i_near[ind] = j2[ind], i2[ind]
        ind = np.where(np.logical_and(in_coords[:,0]>=0.5, in_coords[:,1]>=0.5))
        j_near[ind], i_near[ind] = j3[ind], i3[ind]
        ind = np.where(np.logical_and(in_coords[:,0]<0.5, in_coords[:,1]>=0.5))
        j_near[ind], i_near[ind] = j4[ind], i4[ind]
        nearest = j_near * self.nx + i_near
        
        return inside, indices, vertices, in_coords, nearest

    def _fill_pole_void(self, fld):
        """
        if rotation of vectors (or other reasons) generates nan at the poles
        we fill in the void using surrounding values for each pole defined by self.pole_dim and pole_index
        """
        if self.pole_dim == 'x':
            for i in self.pole_index:
                if i==0:
                    fld[:, 0] = np.mean(fld[:, 1])
                if i==-1:
                    fld[:, -1] = np.mean(fld[:, -2])
        if self.pole_dim == 'y':
            for i in self.pole_index:
                if i==0:
                    fld[0, :] = np.mean(fld[1, :])
                if i==-1:
                    fld[-1, :] = np.mean(fld[-2, :])
        return fld

    def rotate_vectors(self, vec_fld):
        vec_fld = super().rotate_vectors(vec_fld)
        for i in range(2):
            vec_fld[i,...] = self._fill_pole_void(vec_fld[i,...])
        return vec_fld

    def get_corners(self, fld):
        """
        given fld defined on a regular grid, obtain its value on the 4 corners/vertices
        """
        assert fld.shape == self.x.shape, "fld shape does not match x,y"
        nx, ny = fld.shape
        fld_ = np.zeros((nx+1, ny+1))
        ##use linear interp in interior
        fld_[1:nx, 1:ny] = 0.25*(fld[1:nx, 1:ny] + fld[1:nx, 0:ny-1] + fld[0:nx-1, 1:ny] + fld[0:nx-1, 0:ny-1])
        ##use 2nd-order polynomial extrapolat along borders
        fld_[0, :] = 3*fld_[1, :] - 3*fld_[2, :] + fld_[3, :]
        fld_[nx, :] = 3*fld_[nx-1, :] - 3*fld_[nx-2, :] + fld_[nx-3, :]
        fld_[:, 0] = 3*fld_[:, 1] - 3*fld_[:, 2] + fld_[:, 3]
        fld_[:, ny] = 3*fld_[:, ny-1] - 3*fld_[:, ny-2] + fld_[:, ny-3]
        ##make corners into new dimension
        fld_corners = np.zeros((nx, ny, 4))
        fld_corners[:, :, 0] = fld_[0:nx, 0:ny]
        fld_corners[:, :, 1] = fld_[0:nx, 1:ny+1]
        fld_corners[:, :, 2] = fld_[1:nx+1, 1:ny+1]
        fld_corners[:, :, 3] = fld_[1:nx+1, 0:ny]
        return fld_corners

    def _interp_weights(self, inside, vertices, in_coords):
        """
        Compute interpolation weights from the outputs of find_index
        the interp_weights are the weights (sums to 1) given to each grid vertex in self.x,y
        based on their distance to the x_,y_ points (as specified by the in_coords)

        Args:
            inside, vertices, in_coords: from the output of self.find_index

        Returns:
            interp_weights (np.array): interpolation weights
        """
        ##compute bilinear interp weights
        interp_weights = np.zeros(vertices.shape)
        interp_weights[:, 0] =  (1-in_coords[:, 0]) * (1-in_coords[:, 1])
        interp_weights[:, 1] =  in_coords[:, 0] * (1-in_coords[:, 1])
        interp_weights[:, 2] =  in_coords[:, 0] * in_coords[:, 1]
        interp_weights[:, 3] =  (1-in_coords[:, 0]) * in_coords[:, 1]
        return interp_weights

    def interp(self, fld, x=None, y=None, method='linear'):
        if x is None or y is None:
            ##use precalculated weights for self.dst_grid
            inside = self.interp_inside
            indices = self.interp_indices
            vertices = self.interp_vertices
            nearest = self.interp_nearest
            weights = self.interp_weights
            x = self.dst_grid.x
        else:
            ##otherwise compute the weights for the given x,y
            inside, indices, vertices, in_coords, nearest = self.find_index(x, y)
            weights = self._interp_weights(inside, vertices, in_coords)

        fld_interp = np.full(np.array(x).flatten().shape, np.nan)
        if fld.shape == self.x.shape:
            if method == 'nearest':
                # find the node of the triangle with the maximum weight
                fld_interp[inside] = fld.flatten()[nearest]
            elif method == 'linear':
                # sum over the weights for each node of triangle
                fld_interp[inside] = np.einsum('nj,nj->n', np.take(fld.flatten(), vertices), weights)
            else:
                raise NotImplementedError(f"interp method {method} is not yet available")
        else:
            raise ValueError(f"field shape {fld.shape} does not match grid shape {self.x.shape}")
        return fld_interp.reshape(np.array(x).shape)

    ###utility function for coarse-graining (high->low resolution)
    def coarsen(self, fld):
        ##find which location x_,y_ falls in in dst_grid
        if fld.shape == self.x.shape:
            inside = self.coarsen_inside
            nearest = self.coarsen_nearest
        else:
            raise ValueError(f"field shape {fld.shape} does not match grid shape {self.x.shape}")

        fld_coarse = np.zeros(self.dst_grid.x.flatten().shape)
        count = np.zeros(self.dst_grid.x.flatten().shape)
        fld_inside = fld.flatten()[inside]
        valid = ~np.isnan(fld_inside)  ##filter out nan

        ##average the fld points inside each dst_grid box
        np.add.at(fld_coarse, nearest[valid], fld_inside[valid])
        np.add.at(count, nearest[valid], 1)

        valid = (count>1)  ##do not coarse grain if only one point near by
        fld_coarse[valid] /= count[valid]
        fld_coarse[~valid] = np.nan

        return fld_coarse.reshape(self.dst_grid.x.shape)

    def plot_field(self, ax, fld,  vmin=None, vmax=None, cmap='viridis', **kwargs):
        if vmin is None:
            vmin = np.nanmin(fld)
        if vmax is None:
            vmax = np.nanmax(fld)

        if isinstance(cmap, str):
            cmap = colormaps[cmap]

        x = self.x
        y = self.y
        ##in case of lon convention 0:360, need to reorder so that x is monotonic
        if self.proj_name == 'longlat':
            ind = np.argsort(x[0,:])
            x = np.take(x, ind, axis=1)
            fld = np.take(fld, ind, axis=1)
        im = ax.pcolor(x, y, fld, vmin=vmin, vmax=vmax, cmap=cmap, **kwargs)

        self.set_xylim(ax)
        return im



from abc import ABC, abstractmethod
import numpy as np
from NEDAS.config import Config

class Transform(ABC):
    """
    Base class for miscellaneous transform functions
    """
    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def forward_state(self, c: Config, rec: dict, field: np.ndarray) -> np.ndarray:
        """
        Forward transform for the model state variables.

        Args:
            c (Config): Configuration object.
            rec (dict): State information record for the variable.
            field (np.ndarray): The state variable.

        Returns:
            np.ndarray: The transformed state variable.
        """
        return field

    @abstractmethod
    def backward_state(self, c: Config, rec: dict, field: np.ndarray) -> np.ndarray:
        """
        Backward (inverse) transform for the model state variables.

        Args:
            c (Config): Configuration object.
            rec (dict): State information record for the variable.
            field (np.ndarray): The transformed state variable.

        Returns:
            np.ndarray: The state variable transformed back to the original space.
        """
        return field

    @abstractmethod
    def forward_obs(self, c: Config, obs_rec: dict, obs_seq: dict) -> dict:
        """
        Forward transform for the observation sequence.

        Args:
            c (Config): Configuration object.
            obs_rec (dict): Observation information record.
            obs_seq (dict): The observation sequence.
                With keys ``'obs'`` the observation;
                ``'x', 'y', 'z', 't'`` the space/time coordinates;
                and ``'err_std'`` the observation errors.

        Returns:
            dict: The transformed observation sequence.
        """
        return obs_seq

    @abstractmethod
    def backward_obs(self, c: Config, obs_rec: dict, obs_seq: dict) -> dict:
        """
        Backward (inverse) transform for the observation sequence.

        Args:
            c (Config): Configuration object.
            obs_rec (dict): Observation information record.
            obs_seq (dict): The transformed observation sequence.
                With keys ``'obs'`` the observation;
                ``'x', 'y', 'z', 't'`` the space/time coordinates;
                and ``'err_std'`` the observation errors.

        Returns:
            dict: The observation sequence transformed back to the original space.
        """
        return obs_seq




import os
import glob
import numpy as np
from datetime import datetime, timedelta
import pyproj
from scipy.spatial import KDTree
from NEDAS.grid import Grid
from NEDAS.datasets import Dataset
from .utils import get_data_traj_pairs, get_triangulation, get_velocity, get_velocity_gradients, get_deform_div, get_deform_shear, get_deform_vort

class RgpsObs(Dataset):

    def __init__(self, config_file=None, parse_args=False, **kwargs):
        super().__init__(config_file, parse_args, **kwargs)

        ##variable dictionary for RGPS naming convention
        self.variables = {
            'seaice_drift': {'dtype':'float', 'is_vector':True, 'z_units':'m', 'units':'km/day'},
            'seaice_deform_shear': {'dtype':'float', 'is_vector':False, 'z_units':'m', 'units':'1/day'},
            'seaice_deform_div': {'dtype':'float', 'is_vector':False, 'z_units':'m', 'units':'1/day'},
            'seaice_deform_vort': {'dtype':'float', 'is_vector':False, 'z_units':'m', 'units':'1/day'},
            }

        ##RGPS trajectory data (x,y) is in NorthPolarStereo projection:
        self.proj = pyproj.Proj(self.proj4)

        ##tolerance when searching for the time along trajectory
        self.dt_tol=timedelta(days=self.DAYS_SEARCH_TOLERANCE)

        self.obs_operator = {}
        ##TODO the get_traj_pairs functions sometimes fail, near the tree.query part, returns wrong index?
        ## for now use the model.diag_variables for deform variables instead
        #    'seaice_drift': self.get_seaice_drift,
        #    'seaice_deform_shear': self.get_seaice_deform_shear,
        #    'seaice_deform_div': self.get_seaice_deform_div,
        #    'seaice_deform_vort': self.get_seaice_deform_vort,
        #}

    def parse_kwargs(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        assert kwargs['time'] is not None, 'rgps.read_obs: time is not defined'
        assert kwargs['obs_window_max'] is not None, 'rgps.read_obs: obs_window_max is not defined'
        assert kwargs['obs_window_min'] is not None, 'rgps.read_obs: obs_window_min is not defined'
        return kwargs

    def filename(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        t = kwargs['time']

        if 'obs_window_min' in kwargs and 'obs_window_max' in kwargs:
            d_range = [kwargs['obs_window_min'], kwargs['obs_window_max']]
        else:
            d_range = [0]

        file_list = []
        search = os.path.join(kwargs['path'], 'RGPS*traj.mat')
        for result in glob.glob(search):
            ss = result.split('_')
            t1 = datetime.strptime(ss[1], '%Y-%m-%d')
            t2 = datetime.strptime(ss[2], '%Y-%m-%d')
            for d in d_range:
                t_ = t + d * timedelta(hours=1)
                if t_ >= t1 and t_ <= t2 and result not in file_list:
                    file_list.append(result)

        assert len(file_list)>0, 'no matching files found'

        return file_list

    def read_obs(self, **kwargs):
        """read obs from rgps dataset"""
        kwargs = super().parse_kwargs(**kwargs)
        obs_name = kwargs['name']
        d0_out = kwargs['time'] + timedelta(hours=1) * kwargs['obs_window_min']
        d1_out = kwargs['time'] + timedelta(hours=1) * kwargs['obs_window_max']

        ##target grid for obs_seq
        grid = kwargs['grid']

        ##note: x,y are obs location on grid.proj (in meters)
        ##      x0,y0,x1,y1 are position on rgps_proj (in kilometers),
        ##      triangles: velocity is defined on nodes and deform on elements
        ##      record: trajectory id, we process one record at a time
        obs_seq = {'obs':[], 'err_std':[], 't':[], 'z':[], 'y':[], 'x':[],
                   'x0':[], 'y0':[], 'triangles':[], 'record':[], 'index':[]}

        rec = 0
        for file_name in self.filename(**kwargs):
            pairs = get_data_traj_pairs(file_name, d0_out, d1_out, self.dt_tol)

            for x0, y0, t0, x1, y1, t1 in pairs:
                tri = get_triangulation(x0, y0)
                u, v = get_velocity(x0, y0, t0, x1, y1, t1)

                ##convert x0,y0 from rgps_proj to grid.proj in meters
                x, y = grid.proj(*self.proj(x0*1000, y0*1000, inverse=True))

                if obs_name == 'seaice_drift':
                    for p in range(len(x)):
                        ##quality check
                        if np.hypot(u[p], v[p]) > self.DRIFT_MAX or np.isnan(u[p]) or np.isnan(v[p]):
                            continue

                        ##add the obs to sequence
                        obs_seq['obs'].append([u[p], v[p]])
                        obs_seq['err_std'].append(self.DRIFT_ERR_STD)
                        obs_seq['t'].append(t0[p])
                        obs_seq['z'].append(0.)
                        obs_seq['y'].append(y[p])
                        obs_seq['x'].append(x[p])
                        obs_seq['record'].append(rec)
                        obs_seq['index'].append(p)

                elif obs_name in ['seaice_deform_shear', 'seaice_deform_div', 'seaice_deform_vort']:
                    x_elem = np.mean(x[tri.triangles], axis=1)
                    y_elem = np.mean(y[tri.triangles], axis=1)
                    t_elem = t0[tri.triangles[:, 0]]
                    ux, uy, vx, vy = get_velocity_gradients(x0, y0, u, v)

                    if obs_name == 'seaice_deform_shear':
                        obs_values = np.hypot(ux - vy, uy + vx)
                    elif obs_name == 'seaice_deform_div':
                        obs_values = ux + vy
                    elif obs_name == 'seaice_deform_vort':
                        obs_values = vx - uy

                    for p in range(len(x_elem)):
                        ##quality check
                        if tri.mask[p]:
                            continue
                        if np.abs(obs_values[p]) > self.DEFORM_MAX or np.isnan(obs_values[p]):
                            continue

                        ##add the obs to sequence
                        obs_seq['obs'].append(obs_values[p])
                        obs_seq['err_std'].append(self.DEFORM_ERR_STD)
                        obs_seq['t'].append(t_elem[p])
                        obs_seq['z'].append(0.)
                        obs_seq['y'].append(y_elem[p])
                        obs_seq['x'].append(x_elem[p])
                        obs_seq['record'].append(rec)
                        obs_seq['index'].append(p)

                ##save raw data for state_to_obs to use
                obs_seq['x0'].append(x0)
                obs_seq['y0'].append(y0)
                obs_seq['triangles'].append(tri.triangles)
                rec += 1

        ##convert from list to np.array
        ##raw data are kept in list format
        for key in ('obs', 'err_std', 't', 'y', 'x', 'z'):
            obs_seq[key] = np.array(obs_seq[key])

        if self.variables[obs_name]['is_vector']:
            obs_seq['obs'] = obs_seq['obs'].T

        ##superob to target grid, coarsen and keep the valid grid points
        ##make a mesh grid for the obs points
        obs_x, obs_y = obs_seq['x'], obs_seq['y']
        obs_grid = Grid(grid.proj, obs_x, obs_y, regular=False)
        ##remove unwanted triangles in the mesh
        msk = np.logical_or(obs_grid.tri.a > 2e8, obs_grid.tri.p > 1e5, obs_grid.tri.ratio < 0.3)
        obs_grid = Grid(grid.proj, obs_x, obs_y, regular=False, triangles=obs_grid.tri.triangles[~msk,:])
        ##convert to target grid with coarse graining (superobing)
        obs_grid.set_destination_grid(grid)
        obs_on_grid = obs_grid.convert(obs_seq['obs'], is_vector=self.variables[obs_name]['is_vector'], coarse_grain=True)
        ##overwrite the obs info with superobs
        if self.variables[obs_name]['is_vector']:
            msk = np.isnan(obs_on_grid[0,...])
            obs_seq['obs'] = np.array([obs_on_grid[0,~msk].flatten(), obs_on_grid[1,~msk].flatten()])
        else:
            msk = np.isnan(obs_on_grid)
            obs_seq['obs'] = obs_on_grid[~msk].flatten()
        obs_seq['x'] = grid.x[~msk].flatten()
        obs_seq['y'] = grid.y[~msk].flatten()
        ##other parameters
        for key in ('z', 't', 'err_std'):
            obs_seq[key] = np.full(obs_seq['x'].size, obs_seq[key][0])

        return obs_seq

    def random_network(self):
        raise NotImplementedError

    def get_model_files(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)

        ##start and end time of the trajectories
        t0 = kwargs['time'] + timedelta(hours=1) * kwargs['obs_window_min']
        t1 = kwargs['time'] + timedelta(hours=1) * kwargs['obs_window_max']

        if kwargs['member'] is not None:
            mstr = '{:03d}'.format(kwargs['member']+1)
        else:
            mstr = ''

        file_list = []
        t_list = []
        search = os.path.join(kwargs['path'], '..', '..', '*', kwargs['model_src'], mstr, 'mesh*.bin')
        for result in glob.glob(search):
            if '00Z' not in result:  ##discard files without time string
                continue
            tstr = result.split('.')[-2].split('_')[-1]
            if 'post_regrid' in result:
                tstr = tstr.replace('0Z', '1Z')
            t = datetime.strptime(tstr, '%Y%m%dT%H%M%SZ')
            if t<t0 or t>t1:
                continue
            file_list.append(result)
            t_list.append(t)

        file_list = [file_list[n] for n in np.argsort(t_list)]
        assert len(file_list)>0, 'no matching files found'

        return file_list

    def get_traj_pairs(self, **kwargs):
        """get nextsim simulated trajectory pairs, corresponding to rgps"""
        model = kwargs['model']

        ##rgps start positions on rgps_proj, in km units
        x0, y0 = kwargs['x0'], kwargs['y0']

        ##make a copy of start position to be updated to final position
        nrec = len(x0)
        x, y, i, dx, dy = [], [], [], [], []
        for r in range(nrec):
            x.append(x0[r].copy()*1000)
            y.append(y0[r].copy()*1000)
            i.append(np.zeros(x0[r].shape, dtype='int'))
            dx.append(np.zeros(x0[r].shape))
            dy.append(np.zeros(x0[r].shape))

        ##time for the start/end position
        t0 = kwargs['time'] + timedelta(hours=1) * kwargs['obs_window_min']
        t1 = kwargs['time'] + timedelta(hours=1) * kwargs['obs_window_max']

        file_list = self.get_model_files(**kwargs)

        for n in range(len(file_list)):
            fname = file_list[n]

            ##get model mesh points in native proj, in meters
            model.read_grid(meshfile=fname, **kwargs)
            mx = model.grid.x
            my = model.grid.y
            ##convert to rgps_proj, in km units
            x_, y_ = self.proj(*model.grid.proj(mx, my, inverse=True))

            if n==0 or 'post_regrid' in fname:
                ##build kdtree for spatial search if model mesh changed
                tree = KDTree(np.vstack([x_, y_]).T)

                ##search for traj position in mesh
                for r in range(nrec):
                    d, i[r] = tree.query(np.vstack([x[r], y[r]]).T)
                    dx[r], dy[r] = x[r] - x_[i[r]], y[r] - y_[i[r]]

            if 'post_regrid' not in fname:
                ##move to next position
                for r in range(nrec):
                    x[r], y[r] = x_[i[r]] + dx[r], y_[i[r]] + dy[r]

        pairs = []
        for r in range(nrec):
            pairs.append([x0[r], y0[r], np.full(x0[r].shape, t0),
                          x[r]/1000,  y[r]/1000, np.full(x0[r].shape, t1)])
        return pairs

    def _get_seaice_property(self, obs_name, compute_func, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        pairs = self.get_traj_pairs(**kwargs)
        xo, yo, tri, r, i = kwargs['x'], kwargs['y'], kwargs['triangles'], kwargs['record'], kwargs['index']
        nobs = xo.size

        if self.variables[obs_name]['is_vector']:
            obs_seq = np.full((2, nobs), np.nan)
        else:
            obs_seq = np.full(nobs, np.nan)

        rec = 0
        for x0, y0, t0, x1, y1, t1 in pairs:
            rind = np.where(r==rec)  ##indices in xo for record rec
            ind = i[rind]            ##indices in x0 that forms the record rec part of xo
            obs_value = compute_func(x0, y0, t0, x1, y1, t1)

            if self.variables[obs_name]['is_vector']:
                obs_seq[:, rind] = obs_value[:, ind]
            else:
                obs_seq[rind] = obs_value[ind]
            rec += 1
        return obs_seq

    def get_seaice_drift(self, **kwargs):
        return self._get_seaice_property('seaice_drift', get_velocity, **kwargs)

    def get_seaice_deform_shear(self, **kwargs):
        return self._get_seaice_property('seaice_deform_shear', get_deform_shear, **kwargs)

    def get_seaice_deform_div(self, **kwargs):
        return self._get_seaice_property('seaice_deform_div', get_deform_div, **kwargs)

    def get_seaice_deform_vort(self, **kwargs):
        return self._get_seaice_property('seaice_deform_vort', get_deform_vort, **kwargs)



import os
import argparse
from datetime import datetime, timedelta
import numpy as np
from NEDAS.utils.netcdf_lib import nc_write_var
from NEDAS.models.topaz.abfile import ABFileForcing
from NEDAS.models.topaz.time_format import dayfor
from NEDAS.models.topaz.v5 import Model
from NEDAS.datasets.ecmwf.forecast import Dataset

##default output path
forcing_path = "/cluster/work/users/yingyue/data/ecmwf_fcsts"

def fill_missing(var):
    ##there will be a pole hole after converting from reduced_gg to topaz_grid
    ##just fill it with surrounding values
    j, i = np.where(np.isnan(var))
    d = 2
    for n in range(len(j)):
        var[j[n], i[n]] = np.nanmean(var[j[n]-d:j[n]+d, i[n]-d:i[n]+d])
    return var

def output_abfile(f, filename, name, t, dt, var, units):
    ##make sure directory exists
    os.system("mkdir -p "+os.path.dirname(filename))
    print(f"output to {filename}.a")

    idm = topaz.grid.nx
    jdm = topaz.grid.ny
    cline1 = "ecmwf"
    cline2 = f"{name} ({units})"
    dtime1 = dayfor(topaz.yrflag, t.year, int(t.strftime('%j')), t.hour)
    rdtime = dt / 24

    ##open the file handle
    if f[name] is None:
        f[name] = ABFileForcing(filename, 'w', idm=idm, jdm=jdm, cline1=cline1, cline2=cline2)

    ##write var to file
    f[name].write_field(var, None, name, dtime1, rdtime)

def output_ncfile(filename, name, member, t, dt, tstart, var):
    print("output to "+filename)
    ny, nx = var.shape
    t_step = int((t - tstart) / (dt*timedelta(hours=1)))
    nc_write_var(filename, {'time':None, 'y':ny, 'x':nx}, name, var, recno={'time':t_step})

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('time_start', help="Start time: format 'ccyymmddHHMM'")
    parser.add_argument('forecast_hours', help="Forecast hours")
    parser.add_argument('-m', '--member', default=0, help=f"Ensemble member to be processed (default: 0)")
    parser.add_argument('-o', '--output', default=forcing_path, help=f"Output directory (default: {forcing_path})")
    args = parser.parse_args()

    time_start = datetime.strptime(args.time_start, '%Y%m%d%H%M')
    time_end = time_start + int(args.forecast_hours) * timedelta(hours=1)
    member = int(args.member)
    forcing_path = args.output

    ecmwf = Dataset(time_start=time_start)
    dt = ecmwf.dt_hours
    topaz = Model()
    variables = topaz.atmos_forcing_variables

    file_list = []
    f = {}
    for name in topaz.force_synoptic_names:
        f[name] = None

    t = time_start
    while t <= time_end:

        for varname, rec in variables.items():
            print(f"\nprocess forcing variable '{varname}' at {t} for member {member}")

            ecmwf.read_grid(name=varname, time=t)
            var = ecmwf.read_var(name=varname, member=member, time=t, units=rec['units'])

            print("convert to topaz grid")
            ecmwf.grid.set_destination_grid(topaz.grid)
            var_topaz = ecmwf.grid.convert(var, is_vector=rec['is_vector'])

            path = os.path.join(forcing_path, f"{ecmwf.time_start:%Y%m%d%H%M}")
            if member == 0:
                forcing_file = topaz.filename(path=path, name=varname, time=t)
                forcing_file_nc = os.path.join(path, "forcing.nc")
            else:
                forcing_file = topaz.filename(path=path, name=varname, member=member-1, time=t)
                forcing_file_nc = os.path.join(path, f"forcing_mem{member:03d}.nc")
            if rec['is_vector']:
                for i in range(2):
                    fill_missing(var_topaz[i,...])
                    output_abfile(f, forcing_file+'.'+rec['name'][i], rec['name'][i], t, dt, var_topaz[i,...], rec['units'])
                    output_ncfile(forcing_file_nc, rec['name'][i], member, t, dt, time_start, var_topaz[i,...])
            else:
                fill_missing(var_topaz)
                output_abfile(f, forcing_file+'.'+rec['name'], rec['name'], t, dt, var_topaz, rec['units'])
                output_ncfile(forcing_file_nc, rec['name'], member, t, dt, time_start, var_topaz)

        t += dt * timedelta(hours=1)

    for name in topaz.force_synoptic_names:
        if f[name] is not None:
            f[name].close()
